from parsons.databases.mysql.mysql import MySQL, MySQLTable
from parsons.etl.table import Table
from test.utils import assert_matching_tables
import unittest
import os


# These tests interact directly with the MySQL database. To run, set env variable "LIVE_TEST=True"
@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestMySQLLive(unittest.TestCase):

    def setUp(self):

        self.mysql = MySQL()

    def tearDown(self):

        # Drop the view, the table and the schema
        sql = f"DROP TABLE IF EXISTS test;"
        self.mysql.query(sql)

    def test_query(self):

        # Check that query sending back expected result
        r = self.mysql.query("SELECT 1")
        self.assertEqual(r.first, 1)

    def test_query_no_response(self):

        # Check that a query that has no response and doesn't fail
        sql = f"CREATE TABLE test (name VARCHAR(255), user_name VARCHAR(255))"
        r = self.mysql.query(sql)
        self.assertEqual(r, None)

    def test_insert_data(self):

        sql = "CREATE TABLE test (name VARCHAR(255), user_name VARCHAR(255));"
        self.mysql.query(sql)

        sql = "INSERT INTO test (name, user_name) VALUES ('me', 'myuser');"
        self.mysql.query(sql)

        r = self.mysql.query("select * from test")

        assert_matching_tables(Table([{'name': 'me', 'user_name': 'myuser'}]), r)


# These tests interact directly with the MySQL database. To run, set env variable "LIVE_TEST=True"
@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestMySQL(unittest.TestCase):

    def setUp(self):

        self.mysql = MySQL()

        # Create tables
        self.mysql.query("CREATE TABLE IF NOT EXISTS test (name VARCHAR(255), user_name VARCHAR(255), id INT)")
        self.mysql.query("""
                         INSERT INTO test (name, user_name, id) 
                         VALUES ('me', 'myuser', '1'),
                                ('you', 'hey', '2'),
                                ('you', 'hey', '3')
                         """)

        self.tbl = MySQLTable(self.mysql, 'test')

    def tearDown(self):

        self.mysql.query("DROP TABLE IF EXISTS test;")

    def test_num_rows(self):

        self.assertEqual(self.tbl.num_rows, 3)

    def test_max_primary_key(self):

        self.assertEqual(self.tbl.max_primary_key('id'), 3)

    def test_distinct_primary_key(self):

        self.assertTrue(self.tbl.distinct_primary_key('id'))
        self.assertFalse(self.tbl.distinct_primary_key('user_name'))

    def test_columns(self):

        self.assertEqual(self.tbl.columns, ['name', 'user_name', 'id'])

    def test_exists(self):

        self.assertTrue(self.tbl.exists)

        tbl_bad = MySQLTable(self.mysql, 'bad_test')
        self.assertFalse(tbl_bad.exists)

    def test_drop(self):

        self.tbl.drop()
        self.assertFalse(self.tbl.exists)

    def test_truncate(self):

        self.tbl.truncate()
        self.assertEqual(self.tbl.num_rows, 0)

    def test_get_rows(self):

        data = [['name', 'user_name', 'id'],
                ['me', 'myuser', '1'],
                ['you', 'hey', '2'],
                ['you', 'hey', '3']]
        tbl = Table(data)

        assert_matching_tables(self.tbl.get_rows(), tbl)

    def test_get_new_rows(self):

        data = [['name', 'user_name', 'id'],
                ['you', 'hey', '2'],
                ['you', 'hey', '3']]
        tbl = Table(data)

        # Basic
        assert_matching_tables(self.tbl.get_new_rows('id', 1), tbl) 

        # Chunking
        assert_matching_tables(self.tbl.get_new_rows('id', 1, chunk_size=1), tbl)

    def test_get_new_rows_count(self):

        self.assertEqual(self.tbl.get_new_rows_count('id', 1), 2)


class TestMySQL(unittest.TestCase):

    def setUp(self):

        self.mysql = MySQL(username='test', password='test', host='test', db='test', port=123)

        self.tbl = Table([['ID', 'Name', 'Score'],
                          [1, 'Jim', 1.9],
                          [2, 'John', -0.5],
                          [3, 'Sarah', .0004]])

    def test_data_type(self):

        # Test smallint
        self.assertEqual(self.mysql.data_type(1, ''), 'smallint')
        # Test int
        self.assertEqual(self.mysql.data_type(32769, ''), 'mediumint')
        # Test bigint
        self.assertEqual(self.mysql.data_type(2147483648, ''), 'bigint')
        # Test varchar that looks like an int
        self.assertEqual(self.mysql.data_type('00001', ''), 'varchar')
        # Test a float as a decimal
        self.assertEqual(self.mysql.data_type(5.001, ''), 'float')
        # Test varchar
        self.assertEqual(self.mysql.data_type('word', ''), 'varchar')
        # Test int with underscore
        self.assertEqual(self.mysql.data_type('1_2', ''), 'varchar')
        # Test int with leading zero
        self.assertEqual(self.mysql.data_type('01', ''), 'varchar')

    def test_evaluate_table(self):

        table_map = [{'name': 'ID', 'type': 'smallint', 'width': 0},
                     {'name': 'Name', 'type': 'varchar', 'width': 10},
                     {'name': 'Score', 'type': 'float', 'width': 0}]
        self.assertEqual(self.mysql.evaluate_table(self.tbl), table_map)

    def test_create_statement(self):

        stmt = "CREATE TABLE test_table ( \n id smallint \n,name varchar(10) \n,score float \n);"
        self.assertEqual(self.mysql.create_statement(self.tbl, 'test_table'), stmt)



from parsons.utilities import check_env
from parsons.utilities.api_connector import APIConnector
import re
from parsons.etl import Table
import logging

logger = logging.getLogger(__name__)

PAGE_SIZE = 100


class Freshdesk():
    """
    Instantiate FreshDesk Class

    `Args:`
        domain:
            The subdomain of the FreshDesk account. Not required if ``FRESHDESK_DOMAIN``
            env variable set.
        api_key:
            The FreshDesk provided application key. Not required if ``FRESHDESK_API_KEY``
            env variable set.
    `Returns:`
        FreshDesk Class
    """

    def __init__(self, domain, api_key):

        self.api_key = check_env.check('FRESHDESK_API_KEY', api_key)
        self.domain = check_env.check('FRESHDESK_DOMAIN', domain)
        self.uri = f'https://{self.domain}.freshdesk.com/api/v2/'
        self.client = APIConnector(self.uri, auth=(self.api_key, 'x'))

    def get_request(self, endpoint, params=None, **kwargs):
        # Internal method to make a get request.

        base_params = {'per_page': PAGE_SIZE}

        if params:
            base_params.update(params)

        r = self.client.request(self.uri + endpoint, 'GET', params=base_params)
        self.client.validate_response(r)
        data = r.json()

        # Paginate
        while 'link' in r.headers.keys():
            logger.info(f'Retrieving another page of {PAGE_SIZE} records.')
            url = re.search('<(.*)>', r.headers['link']).group(1)
            r = self.client.request(url, 'GET', params=params)
            self.client.validate_response(r)
            data.extend(r.json())

        return data

    def transform_table(self, tbl, expand_custom_fields=None):
        # Internal method to transform a table prior to returning
        if tbl.num_rows > 0:
            tbl.move_column('id', 0)
            tbl.sort()
            if expand_custom_fields:
                tbl.unpack_dict('custom_fields', prepend=False)

        return tbl

    def get_tickets(self, ticket_type=None, requester_id=None, requester_email=None,
                    company_id=None, updated_since='2016-01-01', expand_custom_fields=False):
        """
        List tickets.

        .. warning::
            Deleted and Spam tickets are not included. However they can be pulled separately
            by utilizing the ``ticket_type`` parameter.

        .. warning::
            Freshdesk will return a maximum of 9,000 tickets. To access additional tickets,
            utilize the ``updated_since`` parameter.

        `Args:`
            ticket_type: str
                Filter by type of ticket to filter by. Valid fields include ``new_and_my_open``,
                ``watching``, ``spam`` and ``deleted``.
            requester_id: int
                Filter by requester id.
            requester_email: str
                Filter by requester email.
            company_id: int
                Filter by company_id.
            expand_custom_fields: boolean
                Expand nested custom fields to their own columns.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {'filter': ticket_type,
                  'requester_id': requester_id,
                  'requester_email': requester_email,
                  'company_id': company_id,
                  'updated_since': updated_since}

        tbl = Table(self.get_request('tickets', params=params))
        logger.info(f'Found {tbl.num_rows} tickets.')
        return self.transform_table(tbl, expand_custom_fields)

    def get_contacts(self, email=None, mobile=None, phone=None, company_id=None,
                     state=None, updated_since=None, expand_custom_fields=None):
        """
        Get contacts.

        `Args:`
            email: str
                Filter by email address.
            mobile: str
                Filter by mobile phone number.
            phone: str
                Filter by phone number.
            expand_custom_fields: boolean
                Expand nested custom fields to their own columns.
        """

        params = {'email': email,
                  'mobile': mobile,
                  'phone': phone,
                  'company_id': company_id,
                  'state': state,
                  '_updated_since': updated_since}

        tbl = Table(self.get_request('contacts', params=params))
        logger.info(f'Found {tbl.num_rows} contacts.')
        return self.transform_table(tbl, expand_custom_fields)

    def get_companies(self, expand_custom_fields=False):
        """
        List companies.

        `Args:`
            expand_custom_fields: boolean
                Expand nested custom fields to their own columns.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        tbl = Table(self.get_request('companies'))
        logger.info(f'Found {tbl.num_rows} companies.')
        return self.transform_table(tbl, expand_custom_fields)

    def get_agents(self, email=None, mobile=None, phone=None, state=None):
        """
        List agents.

        `Args:`
            email: str
                Filter by email address.
            mobile: str
                Filter by mobile phone number
            phone: str
                Filter by phone number
            state: str
                Filter by state
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {'email': email,
                  'mobile': mobile,
                  'phone': phone,
                  'state': state}

        tbl = Table(self.get_request('agents', params=params))
        logger.info(f'Found {tbl.num_rows} agents.')
        tbl = self.transform_table(tbl)
        tbl = tbl.unpack_dict('contact', prepend=False)
        tbl.remove_column('signature')  # Removing since raw HTML might cause issues.

        return tbl



from parsons.etl import Table
import petl
import censusgeocode
import logging

logger = logging.getLogger(__name__)


# The size of batches to send to the batch geocode endpoint. Currently
# the recommendation is less than 1K records.
BATCH_SIZE = 999


class CensusGeocoder(object):
    """
    Instantiate the CensusGecoder Class

    `Args:`
        benchmark: str
            The US Census benchmark file to utilize. By default the current benchmark is used,
            but other options can found `here <https://geocoding.geo.census.gov/geocoder/benchmarks>`_.
        vintage: str
            The US Census vintage file to utilize. By default the current vintage is used, but
            other options can be found `here <https://geocoding.geo.census.gov/geocoder/vintages?form>`_.
    """ # noqa E501

    def __init__(self, benchmark='Public_AR_Current', vintage='Current_Current'):

        self.cg = censusgeocode.CensusGeocode(benchmark=benchmark, vintage=vintage)

    def geocode_onelineaddress(self, address, return_type='geographies'):
        """
        Geocode a single line address. Does not require parsing of city and zipcode field. Returns
        geocode as well as other census block data. If the service is unable to geocode the address
        it will return an empty list.

        `Args:`
            address: str
                A valid US address
            return_type: str
                ``geographies`` will return information about the Census geographies
                while ``locations`` will information about the address.
        `Returns`:
            dict
        """

        geo = self.cg.onelineaddress(address, returntype=return_type)
        self._log_result(geo)
        return geo

    def geocode_address(self, address_line, city=None, state=None, zipcode=None,
                        return_type='geographies'):
        """
        Geocode an address by specifying address fields. Returns the geocode as well as other
        census block data.

        `Args:`
            address_line: str
                A valid address line
            city: str
                A valid city
            state: str
                A valid two character state abbreviation (e.g. 'IL')
            zipcode: int
                A valid five digit zipcode (e.g. 60622)
            return_type: str
                ``geographies`` will return information about the Census geographies
                while ``locations`` will information about the address.
        `Returns:`
            dict
        """

        geo = self.cg.address(address_line, city=city, state=state, zipcode=zipcode)
        self._log_result(geo)
        return geo

    def geocode_address_batch(self, table):
        """
        Geocode multiple addresses from a parsons table.

        The table must **only** include the following columns in the following order.

        .. list-table::
            :widths: 40
            :header-rows: 1

            * - Column Data
            * - Unique ID
            * - Street
            * - City
            * - State
            * - Zipcode

        `Args:`
            table: Parsons Table
                A Parsons table
        `Returns:`
            A Parsons table
        """

        logger.info(f'Geocoding {table.num_rows} records.')
        chunked_tables = table.chunk(BATCH_SIZE)
        batch_count = 1
        records_processed = 0

        geocoded_tbl = Table([[]])
        for tbl in chunked_tables:
            geocoded_tbl.concat(Table(petl.fromdicts(self.cg.addressbatch(tbl))))
            records_processed += tbl.num_rows
            logger.info(f'{records_processed} of {table.num_rows} records processed.')
            batch_count += 1

        return geocoded_tbl

    def _log_result(self, dict):
        # Internal method to log the result of the geocode

        if len(dict) == 0:
            logger.info('Unable to geocode record.')
        else:
            logger.info('Record geocoded.')

    def get_coordinates_data(self, latitude, longitude):
        """
        Return census data on coordinates.

        `Args`
            latitude: int
                A valid latitude in the United States
            longitude: int
                A valid longitude in the United States
        `Returns:`
           dict
        """

        geo = self.cg.coordinates(x=longitude, y=latitude)
        if len(geo['States']) == 0:
            logger.info('Coordinate not found.')
        else:
            logger.info('Coordinate processed.')
        return geo



from parsons.etl.table import Table
from parsons.databases.redshift.rs_copy_table import RedshiftCopyTable
from parsons.databases.redshift.rs_create_table import RedshiftCreateTable
from parsons.databases.redshift.rs_table_utilities import RedshiftTableUtilities
from parsons.databases.redshift.rs_schema import RedshiftSchema
from parsons.databases.table import BaseTable
from parsons.utilities import files
import psycopg2
import psycopg2.extras
import os
import logging
import json
import pickle
import petl
from contextlib import contextmanager
import datetime
import random

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000

logger = logging.getLogger(__name__)


class Redshift(RedshiftCreateTable, RedshiftCopyTable, RedshiftTableUtilities, RedshiftSchema):
    """
    A Redshift class to connect to database.

    Args:
        username: str
            Required if env variable ``REDSHIFT_USERNAME`` not populated
        password: str
            Required if env variable ``REDSHIFT_PASSWORD`` not populated
        host: str
            Required if env variable ``REDSHIFT_HOST`` not populated
        db: str
            Required if env variable ``REDSHIFT_DB`` not populated
        port: int
            Required if env variable ``REDSHIFT_PORT`` not populated. Port 5439 is typical.
        timeout: int
            Seconds to timeout if connection not established
        s3_temp_bucket: str
            Name of the S3 bucket that will be used for storing data during bulk transfers.
            Required if you intend to perform bulk data transfers (eg. the copy_s3 method),
            and env variable ``S3_TEMP_BUCKET`` is not populated.
        aws_access_key_id: str
            The default AWS access key id for copying data from S3 into Redshift
            when running copy/upsert/etc methods.
            This will default to environment variable AWS_ACCESS_KEY_ID.
        aws_secret_access_key: str
            The default AWS secret access key for copying data from S3 into Redshift
            when running copy/upsert/etc methods.
            This will default to environment variable AWS_SECRET_ACCESS_KEY.
        iam_role: str
            AWS IAM Role ARN string -- an optional, different way for credentials to
            be provided in the Redshift copy command that does not require an access key.
    """

    def __init__(self, username=None, password=None, host=None, db=None, port=None,
                 timeout=10, s3_temp_bucket=None,
                 aws_access_key_id=None, aws_secret_access_key=None, iam_role=None):

        try:
            self.username = username or os.environ['REDSHIFT_USERNAME']
            self.password = password or os.environ['REDSHIFT_PASSWORD']
            self.host = host or os.environ['REDSHIFT_HOST']
            self.db = db or os.environ['REDSHIFT_DB']
            self.port = port or os.environ['REDSHIFT_PORT']
        except KeyError as error:
            logger.error("Connection info missing. Most include as kwarg or "
                         "env variable.")
            raise error

        self.timeout = timeout
        self.dialect = 'redshift'
        self.s3_temp_bucket = s3_temp_bucket or os.environ.get('S3_TEMP_BUCKET')
        # We don't check/load the environment variables for aws_* here
        # because the logic in S3() and rs_copy_table.py does already.
        self.aws_access_key_id = aws_access_key_id
        self.aws_secret_access_key = aws_secret_access_key
        self.iam_role = iam_role

    @contextmanager
    def connection(self):
        """
        Generate a Redshift connection.
        The connection is set up as a python "context manager", so it will be closed
        automatically (and all queries committed) when the connection goes out of scope.

        When using the connection, make sure to put it in a ``with`` block (necessary for
        any context manager):
        ``with rs.connection() as conn:``

        `Returns:`
            Psycopg2 `connection` object
        """

        # Create a psycopg2 connection and cursor
        conn = psycopg2.connect(user=self.username, password=self.password,
                                host=self.host, dbname=self.db, port=self.port,
                                connect_timeout=self.timeout)
        try:
            yield conn

            conn.commit()
        finally:
            conn.close()

    @contextmanager
    def cursor(self, connection):
        cur = connection.cursor(cursor_factory=psycopg2.extras.DictCursor)
        try:
            yield cur
        finally:
            cur.close()

    def query(self, sql, parameters=None):
        """
        Execute a query against the Redshift database. Will return ``None``
        if the query returns zero rows.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the `psycopg style <http://initd.org/psycopg/docs/usage.html#passing-parameters-to-sql-queries>`_.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

            # Note that the name contains a quote, which could break your query if not escaped
            # properly.
            name = "Beatrice O'Brady"
            sql = "SELECT * FROM my_table WHERE name = %s"
            rs.query(sql, parameters=[name])

        .. code-block:: python

            names = ["Allen Smith", "Beatrice O'Brady", "Cathy Thompson"]
            placeholders = ', '.join('%s' for item in names)
            sql = f"SELECT * FROM my_table WHERE name IN ({placeholders})"
            rs.query(sql, parameters=names)

        `Args:`
            sql: str
                A valid SQL statement
            parameters: list
                A list of python variables to be converted into SQL values in your query

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.

        """  # noqa: E501

        with self.connection() as connection:
            return self.query_with_connection(sql, connection, parameters=parameters)

    def query_with_connection(self, sql, connection, parameters=None, commit=True):
        """
        Execute a query against the Redshift database, with an existing connection.
        Useful for batching queries together. Will return ``None`` if the query
        returns zero rows.

        `Args:`
            sql: str
                A valid SQL statement
            connection: obj
                A connection object obtained from ``redshift.connection()``
            parameters: list
                A list of python variables to be converted into SQL values in your query
            commit: boolean
                Whether to commit the transaction immediately. If ``False`` the transaction will
                be committed when the connection goes out of scope and is closed (or you can
                commit manually with ``connection.commit()``).

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        # To Do: Have it return an ordered dict to return the
        #        rows in the correct order

        with self.cursor(connection) as cursor:

            logger.debug(f'SQL Query: {sql}')
            cursor.execute(sql, parameters)

            if commit:
                connection.commit()

            # If the cursor is empty, don't cause an error
            if not cursor.description:
                logger.debug('Query returned 0 rows')
                return None

            else:

                # Fetch the data in batches, and "pickle" the rows to a temp file.
                # (We pickle rather than writing to, say, a CSV, so that we maintain
                # all the type information for each field.)

                temp_file = files.create_temp_file()

                with open(temp_file, 'wb') as f:
                    # Grab the header
                    header = [i[0] for i in cursor.description]
                    pickle.dump(header, f)

                    while True:
                        batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                        if not batch:
                            break

                        logger.debug(f'Fetched {len(batch)} rows.')
                        for row in batch:
                            pickle.dump(list(row), f)

                # Load a Table from the file
                final_tbl = Table(petl.frompickle(temp_file))

                logger.debug(f'Query returned {final_tbl.num_rows} rows.')
                return final_tbl

    def copy_s3(self, table_name, bucket, key, manifest=False, data_type='csv',
                csv_delimiter=',', compression=None, if_exists='fail', max_errors=0,
                distkey=None, sortkey=None, padding=None, varchar_max=None,
                statupdate=True, compupdate=True, ignoreheader=1, acceptanydate=True,
                dateformat='auto', timeformat='auto', emptyasnull=True,
                blanksasnull=True, nullas=None, acceptinvchars=True, truncatecolumns=False,
                columntypes=None, specifycols=None,
                aws_access_key_id=None, aws_secret_access_key=None):
        """
        Copy a file from s3 to Redshift.

        `Args:`
            table_name: str
                The table name and schema (``tmc.cool_table``) to point the file.
            bucket: str
                The s3 bucket where the file or manifest is located.
            key: str
                The key of the file or manifest in the s3 bucket.
            manifest: str
                If using a manifest
            data_type: str
                The data type of the file. Only ``csv`` supported currently.
            csv_delimiter: str
                The delimiter of the ``csv``. Only relevant if data_type is ``csv``.
            compression: str
                If specified (``gzip``), will attempt to decompress the file.
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails.
            distkey: str
                The column name of the distkey
            sortkey: str
                The column name of the sortkey
            padding: float
                A percentage padding to add to varchar columns if creating a new table. This is
                helpful to add a buffer for future copies in which the data might be wider.
            varchar_max: list
                A list of columns in which to set the width of the varchar column to 65,535
                characters.
            statupate: boolean
                Governs automatic computation and refresh of optimizer statistics at the end
                of a successful COPY command.
            compupdate: boolean
                Controls whether compression encodings are automatically applied during a COPY.
            ignore_header: int
                The number of header rows to skip. Ignored if data_type is ``json``.
            acceptanydate: boolean
                Allows any date format, including invalid formats such as 00/00/00 00:00:00, to be
                loaded without generating an error.
            emptyasnull: boolean
                Indicates that Amazon Redshift should load empty char and varchar fields
                as ``NULL``.
            blanksasnull: boolean
                Loads blank varchar fields, which consist of only white space characters,
                as ``NULL``.
            nullas: str
                Loads fields that match string as NULL
            acceptinvchars: boolean
                Enables loading of data into VARCHAR columns even if the data contains
                invalid UTF-8 characters.
            dateformat: str
                Set the date format. Defaults to ``auto``.
            timeformat: str
                Set the time format. Defaults to ``auto``.
            truncatecolumns: boolean
                If the table already exists, truncates data in columns to the appropriate number
                of characters so that it fits the column specification. Applies only to columns
                with a VARCHAR or CHAR data type, and rows 4 MB or less in size.
            columntypes: dict
                Optional map of column name to redshift column type, overriding the usual type
                inference. You only specify the columns you want to override, eg.
                ``columntypes={'phone': 'varchar(12)', 'age': 'int'})``.
            specifycols: boolean
                Adds a column list to the Redshift `COPY` command, allowing for the source table
                in an append to have the columnns out of order, and to have fewer columns with any
                leftover target table columns filled in with the `DEFAULT` value.

                This will fail if all of the source table's columns do not match a column in the
                target table. This will also fail if the target table has an `IDENTITY`
                column and that column name is among the source table's columns.
            aws_access_key_id:
                An AWS access key granted to the bucket where the file is located. Not required
                if keys are stored as environmental variables.
            aws_secret_access_key:
                An AWS secret access key granted to the bucket where the file is located. Not
                required if keys are stored as environmental variables.
        `Returns`
            Parsons Table or ``None``
                See :ref:`parsons-table` for output options.
        """

        with self.connection() as connection:

            if self._create_table_precheck(connection, table_name, if_exists):

                # Grab the object from s3
                from parsons.aws.s3 import S3
                s3 = S3(aws_access_key_id=aws_access_key_id,
                        aws_secret_access_key=aws_secret_access_key)

                local_path = s3.get_file(bucket, key)

                if data_type == 'csv':
                    tbl = Table.from_csv(local_path, delimiter=csv_delimiter)
                else:
                    raise TypeError("Invalid data type provided")

                # Create the table
                sql = self.create_statement(tbl, table_name, padding=padding,
                                            distkey=distkey, sortkey=sortkey,
                                            varchar_max=varchar_max,
                                            columntypes=columntypes)

                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'{table_name} created.')

            # Copy the table
            copy_sql = self.copy_statement(table_name, bucket, key, manifest=manifest,
                                           data_type=data_type, csv_delimiter=csv_delimiter,
                                           compression=compression, max_errors=max_errors,
                                           statupdate=statupdate, compupdate=compupdate,
                                           aws_access_key_id=aws_access_key_id,
                                           aws_secret_access_key=aws_secret_access_key,
                                           ignoreheader=ignoreheader, acceptanydate=acceptanydate,
                                           emptyasnull=emptyasnull, blanksasnull=blanksasnull,
                                           nullas=nullas, acceptinvchars=acceptinvchars,
                                           truncatecolumns=truncatecolumns,
                                           specifycols=specifycols,
                                           dateformat=dateformat, timeformat=timeformat)

            self.query_with_connection(copy_sql, connection, commit=False)
            logger.info(f'Data copied to {table_name}.')

    def copy(self, tbl, table_name, if_exists='fail', max_errors=0, distkey=None,
             sortkey=None, padding=None, statupdate=False, compupdate=True, acceptanydate=True,
             emptyasnull=True, blanksasnull=True, nullas=None, acceptinvchars=True,
             dateformat='auto', timeformat='auto', varchar_max=None, truncatecolumns=False,
             columntypes=None, specifycols=None, alter_table=False,
             aws_access_key_id=None, aws_secret_access_key=None, iam_role=None,
             cleanup_s3_file=True, template_table=None):
        """
        Copy a :ref:`parsons-table` to Redshift.

        `Args:`
            tbl: obj
                A Parsons Table.
            table_name: str
                The destination table name (ex. ``my_schema.my_table``).
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            max_errors: int
                The maximum number of rows that can error and be skipped before
                the job fails.
            distkey: str
                The column name of the distkey
            sortkey: str
                The column name of the sortkey
            padding: float
                A percentage padding to add to varchar columns if creating a new table. This is
                helpful to add a buffer for future copies in which the data might be wider.
            varchar_max: list
                A list of columns in which to set the width of the varchar column to 65,535
                characters.
            statupate: boolean
                Governs automatic computation and refresh of optimizer statistics at the end
                of a successful COPY command.
            compupdate: boolean
                Controls whether compression encodings are automatically applied during a COPY.
            acceptanydate: boolean
                Allows any date format, including invalid formats such as 00/00/00 00:00:00, to be
                loaded without generating an error.
            emptyasnull: boolean
                Indicates that Amazon Redshift should load empty char and varchar fields
                as ``NULL``.
            blanksasnull: boolean
                Loads blank varchar fields, which consist of only white space characters,
                as ``NULL``.
            nullas: str
                Loads fields that match string as NULL
            acceptinvchars: boolean
                Enables loading of data into VARCHAR columns even if the data contains
                invalid UTF-8 characters.
            dateformat: str
                Set the date format. Defaults to ``auto``.
            timeformat: str
                Set the time format. Defaults to ``auto``.
            truncatecolumns: boolean
                If the table already exists, truncates data in columns to the appropriate number
                of characters so that it fits the column specification. Applies only to columns
                with a VARCHAR or CHAR data type, and rows 4 MB or less in size.
            columntypes: dict
                Optional map of column name to redshift column type, overriding the usual type
                inference. You only specify the columns you want to override, eg.
                ``columntypes={'phone': 'varchar(12)', 'age': 'int'})``.
            specifycols: boolean
                Adds a column list to the Redshift `COPY` command, allowing for the source table
                in an append to have the columnns out of order, and to have fewer columns with any
                leftover target table columns filled in with the `DEFAULT` value.

                This will fail if all of the source table's columns do not match a column in the
                target table. This will also fail if the target table has an `IDENTITY`
                column and that column name is among the source table's columns.
            alter_table: boolean
                Will check if the target table varchar widths are wide enough to copy in the
                table data. If not, will attempt to alter the table to make it wide enough. This
                will not work with tables that have dependent views.
            aws_access_key_id:
                An AWS access key granted to the bucket where the file is located. Not required
                if keys are stored as environmental variables.
            aws_secret_access_key:
                An AWS secret access key granted to the bucket where the file is located. Not
                required if keys are stored as environmental variables.
            iam_role: str
                An AWS IAM Role ARN string; an alternative credential for the COPY command
                from Redshift to S3. The IAM role must have been assigned to the Redshift
                instance and have access to the S3 bucket.
            cleanup_s3_file: boolean
                The s3 upload is removed by default on cleanup. You can set to False for debugging.
            template_table: str
                Instead of specifying columns, columntypes, and/or inference, if there
                is a pre-existing table that has the same columns/types, then use the template_table
                table name as the schema for the new table.
                Unless you set specifycols=False explicitly, a template_table will set it to True

        `Returns`
            Parsons Table or ``None``
                See :ref:`parsons-table` for output options.
        """

        # Specify the columns for a copy statement.
        if specifycols or (specifycols is None and template_table):
            cols = tbl.columns
        else:
            cols = None

        with self.connection() as connection:

            # Check to see if the table exists. If it does not or if_exists = drop, then
            # create the new table.
            if self._create_table_precheck(connection, table_name, if_exists):
                if template_table:
                    # Copy the schema from the template table
                    sql = f'CREATE TABLE {table_name} (LIKE {template_table})'
                else:
                    sql = self.create_statement(tbl, table_name, padding=padding,
                                                distkey=distkey, sortkey=sortkey,
                                                varchar_max=varchar_max,
                                                columntypes=columntypes)
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'{table_name} created.')

            # If alter_table is True, then alter table if the table column widths
            # are wider than the existing table.
            elif alter_table:
                self.alter_varchar_column_widths(tbl, table_name)

            # Upload the table to S3
            key = self.temp_s3_copy(tbl, aws_access_key_id=aws_access_key_id,
                                    aws_secret_access_key=aws_secret_access_key)

            try:
                # Copy to Redshift database.
                copy_args = {'max_errors': max_errors,
                             'ignoreheader': 1,
                             'statupdate': statupdate,
                             'compupdate': compupdate,
                             'acceptanydate': acceptanydate,
                             'dateformat': dateformat,
                             'timeformat': timeformat,
                             'blanksasnull': blanksasnull,
                             'nullas': nullas,
                             'emptyasnull': emptyasnull,
                             'acceptinvchars': acceptinvchars,
                             'truncatecolumns': truncatecolumns,
                             'specifycols': cols,
                             'aws_access_key_id': aws_access_key_id,
                             'aws_secret_access_key': aws_secret_access_key,
                             'compression': 'gzip'}

                # Copy from S3 to Redshift
                sql = self.copy_statement(table_name, self.s3_temp_bucket, key, **copy_args)
                logger.debug(f'Copy SQL command: {sql}')
                self.query_with_connection(sql, connection, commit=False)

                logger.info(f'Data copied to {table_name}.')

            # Clean up the S3 bucket.
            finally:
                if key and cleanup_s3_file:
                    self.temp_s3_delete(key)

    def unload(self, sql, bucket, key_prefix, manifest=True, header=True, compression='gzip',
               add_quotes=True, null_as=None, escape=True, allow_overwrite=True,
               parallel=True, max_file_size='6.2 GB', aws_region=None,
               aws_access_key_id=None, aws_secret_access_key=None):
        """
        Unload Redshift data to S3 Bucket. This is a more efficient method than running a query
        to export data as it can export in parallel and directly into an S3 bucket. Consider
        using this for exports of 10MM or more rows.

        sql: str
            The SQL string to execute to generate the data to unload.
        buckey: str
           The destination S3 bucket
        key_prefix: str
            The prefix of the key names that will be written
        manifest: boolean
            Creates a manifest file that explicitly lists details for the data files
            that are created by the UNLOAD process.
        header: boolean
            Adds a header line containing column names at the top of each output file.
        compression: str
            One of ``gzip``, ``bzip2`` or ``None``. Unloads data to one or more compressed
            files per slice. Each resulting file is appended with a ``.gz`` or ``.bz2`` extension.
        add_quotes: boolean
            Places quotation marks around each unloaded data field, so that Amazon Redshift
            can unload data values that contain the delimiter itself.
        null_as: str
            Specifies a string that represents a null value in unload files. If this option is
            not specified, null values are unloaded as zero-length strings for delimited output.
        escape: boolean
            For CHAR and VARCHAR columns in delimited unload files, an escape character (\) is
            placed before every linefeed, carriage return, escape characters and delimiters.
        allow_overwrite: boolean
            If ``True``, will overwrite existing files, including the manifest file. If ``False``
            will fail.
        parallel: boolean
            By default, UNLOAD writes data in parallel to multiple files, according to the number
            of slices in the cluster. The default option is ON or TRUE. If PARALLEL is OFF or
            FALSE, UNLOAD writes to one or more data files serially, sorted absolutely according
            to the ORDER BY clause, if one is used.
        max_file_size: str
            The maximum size of files UNLOAD creates in Amazon S3. Specify a decimal value between
            5 MB and 6.2 GB.
        region: str
            The AWS Region where the target Amazon S3 bucket is located. REGION is required for
            UNLOAD to an Amazon S3 bucket that is not in the same AWS Region as the Amazon Redshift
            cluster.
        aws_access_key_id:
            An AWS access key granted to the bucket where the file is located. Not required
            if keys are stored as environmental variables.
        aws_secret_access_key:
            An AWS secret access key granted to the bucket where the file is located. Not
            required if keys are stored as environmental variables.
        """  # NOQA W605

        statement = f"""
                     UNLOAD ('{sql}') to 's3://{bucket}/{key_prefix}' \n
                     {self.get_creds(aws_access_key_id, aws_secret_access_key)} \n
                     PARALLEL {parallel} \n
                     MAXFILESIZE {max_file_size}
                     """
        if manifest:
            statement += "MANIFEST \n"
        if header:
            statement += "HEADER \n"
        if compression:
            statement += f"{compression.upper()} \n"
        if add_quotes:
            statement += "ADDQUOTES \n"
        if null_as:
            statement += f"NULL {null_as} \n"
        if escape:
            statement += "ESCAPE \n"
        if allow_overwrite:
            statement += "ALLOWOVERWRITE"
        if aws_region:
            statement += f"REGION {aws_region}"

        logger.info(f'Unloading data to s3://{bucket}/{key_prefix}')
        logger.debug(statement)

        return self.query(statement)

    def generate_manifest(self, buckets, aws_access_key_id=None, aws_secret_access_key=None,
                          mandatory=True, prefix=None, manifest_bucket=None, manifest_key=None,
                          path=None):
        """
        Given a list of S3 buckets, generate a manifest file (JSON format). A manifest file
        allows you to copy multiple files into a single table at once. Once the manifest is
        generated, you can pass it with the :func:`~parsons.redshift.Redshift.copy_s3` method.

        AWS keys are not required if ``AWS_ACCESS_KEY_ID`` and
        ``AWS_SECRET_ACCESS_KEY`` environmental variables set.

        `Args:`

            buckets: list or str
                A list of buckets or single bucket from which to generate manifest
            aws_access_key_id: str
                AWS access key id to access S3 bucket
            aws_secret_access_key: str
                AWS secret access key to access S3 bucket
            mandatory: boolean
                The mandatory flag indicates whether the Redshift COPY should
                terminate if the file does not exist.
            prefix: str
                Optional filter for key prefixes
            manifest_bucket: str
                Optional bucket to write manifest file.
            manifest_key: str
                Optional key name for S3 bucket to write file

        `Returns:`
            ``dict`` of manifest
        """

        from parsons.aws import S3
        s3 = S3(aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key)

        # Deal with a single bucket being passed, rather than list.
        if isinstance(buckets, str):
            buckets = [buckets]

        # Generate manifest file
        manifest = {'entries': []}
        for bucket in buckets:

            # Retrieve list of files in bucket
            key_list = s3.list_keys(bucket, prefix=prefix)
            for key in key_list:
                manifest['entries'].append({
                    'url': '/'.join(['s3:/', bucket, key]),
                    'mandatory': mandatory
                })

        logger.info('Manifest generated.')

        # Save the file to s3 bucket if provided
        if manifest_key and manifest_bucket:
            # Dump the manifest to a temp JSON file
            manifest_path = files.create_temp_file()
            with open(manifest_path, 'w') as manifest_file_obj:
                json.dump(manifest, manifest_file_obj, sort_keys=True, indent=4)

            # Upload the file to S3
            s3.put_file(manifest_bucket, manifest_key, manifest_path)

            logger.info(f'Manifest saved to s3://{manifest_bucket}/{manifest_key}')

        return manifest

    def upsert(self, table_obj, target_table, primary_key, vacuum=True, distinct_check=True,
               cleanup_temp_table=True, **copy_args):
        """
        Preform an upsert on an existing table. An upsert is a function in which records
        in a table are updated and inserted at the same time. Unlike other SQL databases,
        it does not exist natively in Redshift.

        `Args:`
            table_obj: obj
                A Parsons table object
            target_table: str
                The schema and table name to upsert
            primary_key: str or list
                The primary key column(s) of the target table
            vacuum: boolean
                Re-sorts rows and reclaims space in the specified table. You must be a table owner
                or super user to effectively vacuum a table, however the method will not fail
                if you lack these priviledges.
            distinct_check: boolean
                Check if the primary key column is distinct. Raise error if not.
            cleanup_temp_table: boolean
                A temp table is dropped by default on cleanup. You can set to False for debugging.
            \**copy_args: kwargs
                See :func:`~parsons.databases.Redshift.copy`` for options.
        """  # noqa: W605

        if not self.table_exists(target_table):
            logger.info('Target table does not exist. Copying into newly \
                         created target table.')
            self.copy(table_obj, target_table)
            return None

        noise = f'{random.randrange(0, 10000):04}'[:4]
        date_stamp = datetime.datetime.now().strftime('%Y%m%d_%H%M')
        # Generate a temp table like "table_tmp_20200210_1230_14212"
        staging_tbl = '{}_stg_{}_{}'.format(target_table, date_stamp, noise)

        if isinstance(primary_key, str):
            primary_keys = [primary_key]
        else:
            primary_keys = primary_key

        if distinct_check:
            primary_keys_statement = ', '.join(primary_keys)
            diff = self.query(f'''
                select (
                    select count(*)
                    from {target_table}
                ) - (
                    SELECT COUNT(*) from (
                        select distinct {primary_keys_statement}
                        from {target_table}
                    )
                ) as total_count
            ''').first
            if diff > 0:
                raise ValueError('Primary key column contains duplicate values.')

        with self.connection() as connection:

            try:
                # Copy to a staging table
                logger.info(f'Building staging table: {staging_tbl}')
                if 'compupdate' not in copy_args:
                    # Especially with a lot of columns, compupdate=True can
                    # cause a lot of processing/analysis by Redshift before upload.
                    # Since this is a temporary table, setting compression for each
                    # column is not impactful barely impactful
                    # https://docs.aws.amazon.com/redshift/latest/dg/c_Loading_tables_auto_compress.html
                    copy_args = dict(copy_args, compupdate=False)
                self.copy(table_obj, staging_tbl,
                          template_table=target_table,
                          **copy_args)

                staging_table_name = staging_tbl.split('.')[1]
                target_table_name = target_table.split('.')[1]

                # Delete rows
                comparisons = [
                    f'{staging_table_name}.{primary_key} = {target_table_name}.{primary_key}'
                    for primary_key in primary_keys
                ]
                where_clause = ' and '.join(comparisons)

                sql = f"""
                       DELETE FROM {target_table}
                       USING {staging_tbl}
                       WHERE {where_clause}
                       """
                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'Target rows deleted from {target_table}.')

                # Insert rows
                # ALTER TABLE APPEND would be more efficient, but you can't run it in a
                # transaction block. It's worth the performance hit to not commit until the
                # end.
                sql = f"""
                       INSERT INTO {target_table}
                       SELECT * FROM {staging_tbl};
                       """

                self.query_with_connection(sql, connection, commit=False)
                logger.info(f'Target rows inserted to {target_table}')

            finally:
                if cleanup_temp_table:
                    # Drop the staging table
                    self.query_with_connection(f"DROP TABLE IF EXISTS {staging_tbl};",
                                               connection, commit=False)
                    logger.info(f'{staging_tbl} staging table dropped.')

        # Vacuum table. You must commit when running this type of transaction.
        if vacuum:
            with self.connection() as connection:
                connection.set_session(autocommit=True)
                self.query_with_connection(f'VACUUM {target_table};', connection)
                logger.info(f'{target_table} vacuumed.')

    def alter_varchar_column_widths(self, tbl, table_name):
        """
        Alter the width of a varchar columns in a Redshift table to match the widths
        of a Parsons table. The columns are matched by column name and not their
        index.

        `Args:`
            tbl: obj
                A Parsons table
            table_name:
                The target table name (e.g. ``my_schema.my_table``)
        `Returns:`
            ``None``
        """

        # Make the Parsons table column names match valid Redshift names
        tbl.table = petl.setheader(tbl.table, self.column_name_validate(tbl.columns))

        # Create a list of column names and max width for string values.
        pc = {c: tbl.get_column_max_width(c) for c in tbl.columns}

        # Determine the max width of the varchar columns in the Redshift table
        s, t = self.split_full_table_name(table_name)
        cols = self.get_columns(s, t)
        rc = {k: v['max_length'] for k, v in cols.items() if v['data_type'] == 'character varying'} # noqa: E501, E261

        # Figure out if any of the destination table varchar columns are smaller than the
        # associated Parsons table columns. If they are, then alter column types to expand
        # their width.
        for c in set(rc.keys()).intersection(set(pc.keys())):
            if rc[c] < pc[c]:
                logger.info(f'{c} not wide enough. Expanding column width.')
                self.alter_table_column_type(table_name, c, 'varchar', varchar_width=pc[c])

    def alter_table_column_type(self, table_name, column_name, data_type, varchar_width=None):
        """
        Alter a column type of an existing table.

        table_name: str
            The table name (ex. ``my_schema.my_table``).
        column_name: str
            The target column name
        data_type: str
            A valid Redshift data type to alter the table to.
        varchar_width:
            The new width of the column if of type varchar.
        """

        sql = f"ALTER TABLE {table_name} ALTER COLUMN {column_name} TYPE {data_type}"

        if varchar_width:
            sql += f"({varchar_width})"

        with self.connection() as connection:
            connection.set_session(autocommit=True)
            self.query_with_connection(sql, connection)
            logger.info(f'Altered {table_name} {column_name}.')

    def table(self, table_name):
        # Return a Redshift table object

        return RedshiftTable(self, table_name)


class RedshiftTable(BaseTable):
    # Redshift table object.

    pass



"""NGPVAN Bulk Import Endpoints"""

import logging
logger = logging.getLogger(__name__)


class BulkImport(object):

    def __init__(self):

        pass

    def get_bulk_import_resources(self):
        """
        Get bulk import resources that available to the user. These define
        the types of bulk imports that you can run. These might include
        ``Contacts``, ``ActivistCodes``, ``ContactsActivistCodes`` and others.

        `Returns:`
            list
                A list of resources.
        """

        r = self.connection.get_request(f'bulkImportJobs/resources')
        logger.info(f'Found {len(r)} bulk import resources.')
        return r



import logging

logger = logging.getLogger(__name__)


class RedshiftTableUtilities(object):

    def __init__(self):
        pass

    def table_exists(self, table_name, view=True):
        """
        Check if a table or view exists in the database.

        `Args:`
            table_name: str
                The table name and schema (e.g. ``myschema.mytable``).
            view: boolean
                Check to see if a view exists by the same name

        `Returns:`
            boolean
                ``True`` if the table exists and ``False`` if it does not.
        """
        with self.connection() as connection:
            return self.table_exists_with_connection(table_name, connection, view)

    def table_exists_with_connection(self, table_name, connection, view=True):
        table_name = table_name.lower().split('.')

        # Check in pg tables for the table
        sql = """select count(*) from pg_tables where schemaname='{}' and
                 tablename='{}';""".format(table_name[0], table_name[1])

        # TODO maybe convert these queries to use self.query_with_connection

        with self.cursor(connection) as cursor:

            cursor.execute(sql)
            result = cursor.fetchone()[0]

            # Check in the pg_views for the table
            if view:
                sql = """select count(*) from pg_views where schemaname='{}' and
                         viewname='{}';""".format(table_name[0], table_name[1])

            cursor.execute(sql)
            result += cursor.fetchone()[0]

        # If in either, return boolean
        if result >= 1:
            logger.info(f'{table_name[0]}.{table_name[1]} exists.')
            return True
        else:
            logger.info(f'{table_name[0]}.{table_name[1]} does NOT exist.')
            return False

    def get_row_count(self, table_name):
        """
        Return the row count of a table.

        **SQL Code**

        .. code-block:: sql

           SELECT COUNT(*) FROM myschema.mytable

        `Args:`
            table_name: str
                The schema and name (e.g. ``myschema.mytable``) of the table.
        `Returns:`
            int
        """

        count_query = self.query(f"select count(*) from {table_name}")
        return count_query[0]['count']

    def rename_table(self, table_name, new_table_name):
        """
        Rename an existing table.

        .. note::
            You cannot move schemas when renaming a table. Instead, utilize
            the :meth:`table_duplicate()`. method.

        Args:
            table_name: str
                Name of existing schema and table (e.g. ``myschema.oldtable``)
            new_table_name: str
                New name for table with the schema omitted (e.g. ``newtable``).
        """

        sql = f"alter table {table_name} rename to {new_table_name}"
        self.query(sql)
        logger.info(f"{table_name} renamed to {new_table_name}")

    def move_table(self, source_table, new_table, drop_source_table=False):
        """
        Move an existing table in the database.It will inherit encoding, sortkey
        and distkey. **Once run, the source table rows will be empty.** This is
        more efficiant than running ``"create newtable as select * from oldtable"``.

        For more information see: `ALTER TABLE APPEND <https://docs.aws.amazon.com/redshift/latest/dg/r_ALTER_TABLE_APPEND.html>`_

        Args:
            source_table: str
                Name of existing schema and table (e.g. ``my_schema.old_table``)
            new_table: str
                New name of schema and table (e.g. ``my_schema.newtable``)
            drop_original: boolean
                Drop the source table.
        Returns:
                None
        """ # noqa: E501,E261

        # To Do: Add the grants
        # To Do: Argument for if the table exists?
        # To Do: Add the ignore extra kwarg.

        create_sql = f"create table {new_table} (like {source_table});"
        alter_sql = f"alter table {new_table} append from {source_table}"

        logger.info(f'Creating empty {new_table} from {source_table}.')
        self.query(create_sql)

        with self.connection() as conn:

            #  An ALTER TABLE statement can't be run within a block, meaning
            #  that it needs to be committed on running. To enable this,
            #  the connection must be set to autocommit.

            conn.set_session(autocommit=True)
            logger.info(f'Moving data from {source_table} to {new_table}.')
            self.query_with_connection(alter_sql, conn)

        if drop_source_table:
            self.query(f"drop table {source_table};")
            logger.info(f'{source_table} dropped.')

        logger.info(f'{source_table} data moved from {new_table}  .')

    def _create_table_precheck(self, connection, table_name, if_exists):
        """
        Helper to determine what to do when you need a table that may already exist.

        `Args:`
            connection: obj
                A connection object obtained from ``redshift.connection()``
            table_name: str
                The table to check
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
        `Returns:`
            bool
                True if the table needs to be created, False otherwise.
        """

        if if_exists not in ['fail', 'truncate', 'append', 'drop']:
            raise ValueError("Invalid value for `if_exists` argument")

        exists = self.table_exists_with_connection(table_name, connection)

        if exists and if_exists in ['fail', 'truncate', 'append']:
            if if_exists == 'fail':
                raise ValueError('Table already exists.')
            elif if_exists == 'truncate':
                truncate_sql = f"truncate table {table_name}"
                self.query_with_connection(truncate_sql, connection, commit=False)

        else:
            if exists and if_exists == 'drop':
                logger.info(f"Table {table_name} exist, will drop...")
                drop_sql = f"drop table {table_name};\n"
                self.query_with_connection(drop_sql, connection, commit=False)

            return True

        return False

    def populate_table_from_query(self, query, destination_table, if_exists='fail', distkey=None,
                                  sortkey=None):
        """
        Populate a Redshift table with the results of a SQL query, creating the table if it
        doesn't yet exist.

        `Args:`
            query: str
                The SQL query
            destination_table: str
                Name of destination schema and table (e.g. ``mys_chema.new_table``)
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
            distkey: str
                The column to use as the distkey for the table.
            sortkey: str
                The column to use as the sortkey for the table.
        """
        with self.connection() as conn:
            should_create = self._create_table_precheck(conn, destination_table, if_exists)

            if should_create:
                logger.info(f"Creating table {destination_table} from query...")
                sql = f"create table {destination_table}"
                if distkey:
                    sql += f" distkey({distkey})"
                if sortkey:
                    sql += f" sortkey({sortkey})"
                sql += f" as {query}"
            else:
                logger.info(f"Inserting data into {destination_table} from query...")
                sql = f"insert into {destination_table} ({query})"

            self.query_with_connection(sql, conn, commit=False)

        logger.info(f'{destination_table} created from query')

    def duplicate_table(self, source_table, destination_table, where_clause='',
                        if_exists='fail', drop_source_table=False):
        """
        Create a copy of an existing table (or subset of rows) in a new
        table. It will inherit encoding, sortkey and distkey.

        `Args:`
            source_table: str
                Name of existing schema and table (e.g. ``myschema.oldtable``)
            destination_table: str
                Name of destination schema and table (e.g. ``myschema.newtable``)
            where_clause: str
                An optional where clause (e.g. ``where org = 1``).
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``,
                or ``truncate`` the table.
            drop_source_table: boolean
                Drop the source table
        """

        with self.connection() as conn:
            should_create = self._create_table_precheck(conn, destination_table, if_exists)

            if should_create:
                logger.info(f'Creating {destination_table} from {source_table}...')
                create_sql = f"create table {destination_table} (like {source_table})"
                self.query_with_connection(create_sql, conn, commit=False)

            logger.info(f"Transferring data to {destination_table} from {source_table}")
            select_sql = f"select * from {source_table} {where_clause}"
            insert_sql = f"insert into {destination_table} ({select_sql})"
            self.query_with_connection(insert_sql, conn, commit=False)

            if drop_source_table:
                logger.info(f'Dropping table {source_table}...')
                drop_sql = f"drop table {source_table}"
                self.query_with_connection(drop_sql, conn, commit=False)

        logger.info(f'{destination_table} created from {source_table}.')

    def union_tables(self, new_table_name, tables, union_all=True, view=False):
        """
        Union a series of table into a new table.

        Args:
            new_table_name: str
                The new table and schema (e.g. ``myschema.newtable``)
            tables: list
                A list of tables to union
            union_all: boolean
                If ``False`` will deduplicate rows. If ``True`` will include
                duplicate rows.
            view: boolean
                Create a view rather than a static table
        Returns:
            None
        """

        union_type = " UNION ALL" if union_all else " UNION"
        table_type = "VIEW" if view else "TABLE"

        sql = f"CREATE {table_type} {new_table_name} AS"
        for index, t in enumerate(tables):
            if index != 0:
                sql += union_type
            sql += f" SELECT * FROM {t}"

        self.query(sql)

        logger.info(f"Created {new_table_name} from {', '.join(tables)}")

    def get_tables(self, schema=None, table_name=None):
        """
        List the tables in a schema including metadata.

        Args:
            schema: str
                Filter by a schema
            table_name: str
                Filter by a table name
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        logger.info(f"Retrieving tables info.")
        sql = "select * from pg_tables"
        if schema or table_name:
            sql += f" where"
        if schema:
            sql += f" schemaname = '{schema}'"
        if table_name:
            if schema:
                sql += " and"
            sql += f" tablename = '{table_name}'"
        return self.query(sql)

    def get_table_stats(self, schema=None, table_name=None):
        """
        List the tables statistics includes row count and size.

        .. warning::
           This method is only accessible by Redshift *superusers*.

        `Args:`
            schema: str
                Filter by a schema
            table_name: str
                Filter by a table name
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        logger.info(f"Retrieving table statistics.")
        sql = "select * from svv_table_info"
        if schema or table_name:
            sql += f" where"
        if schema:
            sql += f" schema = '{schema}'"
        if table_name:
            if schema:
                sql += " and "
            sql += f" \"table\" = '{table_name}'"
        return self.query(sql)

    def get_columns(self, schema, table_name):
        """
        Gets the column names (and some other column info) for a table.

        If you just need the column names, you can treat the return value like a list, eg:

        .. code-block:: python

            for col in rs.get_columns('some_schema', 'some_table'):
                print(col)

        `Args:`
            schema: str
                The schema name
            table_name: str
                The table name
        `Returns:`
            A dict mapping column name to a dict with extra info. The keys of the dict are ordered
            just like the columns in the table. The extra info is a dict with format
            ``{'data_type': str, 'max_length': int or None, 'max_precision': int or None,
               'max_scale': int or None, 'is_nullable': bool}``
        """

        query = f"""
            select ordinal_position,
                   column_name,
                   data_type,
                   character_maximum_length as max_length,
                   numeric_precision as max_precision,
                   numeric_scale as max_scale,
                   is_nullable
            from information_schema.columns
            where table_name = '{table_name}'
            and table_schema = '{schema}'
            order by ordinal_position
        """

        return {
            row['column_name']: {
                'data_type': row['data_type'],
                'max_length': row['max_length'],
                'max_precision': row['max_precision'],
                'max_scale': row['max_scale'],
                'is_nullable': row['is_nullable'] == 'YES',
            }
            for row in self.query(query)
        }

    def get_views(self, schema=None, view=None):
        """
        List views.

        Args:
            schema: str
                Filter by a schema
            view: str
                Filter by a table name
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        logger.info(f"Retrieving views info.")
        sql = """
              select table_schema as schema_name,
              table_name as view_name,
              view_definition
              from information_schema.views
              where table_schema not in ('information_schema', 'pg_catalog')
              """
        if schema:
            sql += f" and table_schema = '{schema}'"
        if view:
            sql += f" and table_name = '{view}'"
        return self.query(sql)

    def get_queries(self):
        """
        Return the Current queries running and queueing, along with resource consumption.

        .. warning::
            Must be a Redshift superuser to run this method.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        logger.info('Retrieving running and queued queries.')

        # Lifted from Redshift Utils https://github.com/awslabs/amazon-redshift-utils/blob/master/src/AdminScripts/running_queues.sql # noqa: E501
        sql = """
              select trim(u.usename) as user,
                s.pid,
                q.xid,
                q.query,
                q.service_class as service_class,
                q.slot_count as slot,
                date_trunc('second',
                q.wlm_start_time) as start,
                decode(trim(q.state),
                    'Running',
                    'Run',
                    'QueuedWaiting',
                    'Queue',
                    'Returning',
                    'Return',trim(q.state)) as state,
                q.queue_Time/1000000 as queue_sec,
                q.exec_time/1000000 as exec_sec,
                m.cpu_time/1000000 cpu_sec,
                m.blocks_read read_mb,
                decode(m.blocks_to_disk,-1,null,m.blocks_to_disk) spill_mb,
                m2.rows as return_rows,
                m3.rows as NL_rows,
                substring(replace(nvl(qrytext_cur.text,trim(translate(s.text,chr(10)||chr(13)||chr(9) ,''))),'\\n',' '),1,90) as sql, -- # noqa: E501
                trim(decode(event&1,1,'SK ','') || decode(event&2,2,'Del ','') || decode(event&4,4,'NL ','') ||  decode(event&8,8,'Dist ','') || decode(event&16,16,'Bcast ','') || decode(event&32,32,'Stats ','')) as Alert -- # noqa: E501
            from stv_wlm_query_state q
            left outer join stl_querytext s on (s.query=q.query and sequence = 0)
            left outer join stv_query_metrics m on ( q.query = m.query and m.segment=-1 and m.step=-1 )
            left outer join stv_query_metrics m2 on ( q.query = m2.query and m2.step_type = 38 )
            left outer join ( select query, sum(rows) as rows from stv_query_metrics m3 where step_type = 15 group by 1) as m3 on ( q.query = m3.query ) -- # noqa: E501
            left outer join pg_user u on ( s.userid = u.usesysid )
            LEFT OUTER JOIN (SELECT ut.xid,'CURSOR ' || TRIM( substring ( TEXT from strpos(upper(TEXT),'SELECT') )) as TEXT
            FROM stl_utilitytext ut
            WHERE sequence = 0
               AND upper(TEXT) like 'DECLARE%'
               GROUP BY text, ut.xid) qrytext_cur ON (q.xid = qrytext_cur.xid)
            left outer join ( select query,sum(decode(trim(split_part(event,':',1)),'Very selective query filter',1,'Scanned a large number of deleted rows',2,'Nested Loop Join in the query plan',4,'Distributed a large number of rows across the network',8,'Broadcasted a large number of rows across the network',16,'Missing query planner statistics',32,0)) as event from STL_ALERT_EVENT_LOG -- # noqa: E501
            where event_time >=  dateadd(hour, -8, current_Date) group by query  ) as alrt on alrt.query = q.query -- # noqa: E501
            """

        return self.query(sql)

    def get_max_value(self, table_name, value_column):
        """
        Return the max value from a table.

        `Args:`
            table_name: str
                Schema and table name
            value_column: str
                The column containing the values
        """

        return self.query(f'SELECT MAX({value_column}) value from {table_name}')[0]['value']

    def get_object_type(self, object_name):
        """
        Get object type.

        One of `view`, `table`, `index`, `sequence`, or `TOAST table`.

        `Args:`
            object_name: str
                The schema.obj for which to get the object type.
        `Returns:`
            `str` of the object type.

        """
        sql_obj_type = f"""
            select n.nspname||'.'||c.relname as objname
            , case
                when relkind='v' then 'view'
                when relkind='r' then 'table'
                when relkind='i' then 'index'
                when relkind='s' then 'sequence'
                when relkind='t' then 'TOAST table'
                end as object_name
            from pg_catalog.pg_class as c
            inner join pg_catalog.pg_namespace as n
                on c.relnamespace = n.oid
            where objname='{object_name}'
        """
        tbl = self.query(sql_obj_type)

        if tbl.num_rows == 0:
            logger.info(f"{object_name} doesn't exist.")
            return None

        return tbl[0]['object_name']

    def is_view(self, object_name):
        """
        Return true if the object is a view.

        `Args:`
            object_name: str
                The schema.obj to test if it's a view.
        `Returns:`
            `bool`

        """
        is_view = self.get_object_type(object_name) == "view"
        logger.info(f"{object_name} is {'' if is_view else 'not'} a view.")
        return is_view

    def is_table(self, object_name):
        """
        Return true if the object is a table.

        `Args:`
            object_name: str
                The schema.obj to test if it's a table.
        `Returns:`
            `bool`

        """
        is_table = self.get_object_type(object_name) == "table"
        logger.info(f"{object_name} is {'' if is_table else 'not'} a table.")
        return is_table

    def get_table_definition(self, table):
        """
        Get the table definition (i.e. the create statement).

        `Args:`
            table: str
                The schema.table for which to get the table definition.
        `Returns:`
            str
        """

        schema, table = self.split_full_table_name(table)

        if not self.is_table(f"{schema}.{table}"):
            return None

        results = self.get_table_definitions(schema, table)

        return results[0]["ddl"]

    def get_table_definitions(self, schema=None, table=None):
        """
        Get the table definition (i.e. the create statement) for multiple tables.

        This works similar to `get_table_def` except it runs a single query
        to get the ddl for multiple tables. It supports SQL wildcards for
        `schema` and `table`. Only returns the ddl for _tables_ that match
        `schema` and `table` if they exist.

        `Args:`
            schema: str
                The schema to filter by.
            table: str
                The table to filter by.
        `Returns:`
            `list` of dicts with matching tables.

        """

        conditions = []
        if schema:
            conditions.append(f"schemaname like '{schema}'")
        if table:
            conditions.append(f"tablename like '{table}'")

        conditions_str = ' and '.join(conditions)
        where_clause = f"where {conditions_str}" if conditions_str else ''

        sql_get_ddl = f"""
            select *
            from admin.v_generate_tbl_ddl
            {where_clause}
        """
        ddl_table = self.query(sql_get_ddl)

        if ddl_table.num_rows == 0:
            logger.info(f"No tables matching {schema} and {table}.")
            return None

        def join_sql_parts(columns, rows):
            return [f"{columns[1]}.{columns[2]}",
                    '\n'.join([row[4] for row in rows])]

        # The query returns the sql over multiple rows
        # We need to join then into a single row
        ddl_table.reduce_rows(
            ['table_id', 'schemaname', 'tablename'],
            join_sql_parts,
            ['tablename', 'ddl'],
            presorted=True)

        return ddl_table.to_dicts()

    def get_view_definition(self, view):
        """
        Get the view definition (i.e. the create statement).

        `Args:`
            view: str
                The schema.view for which to get the view definition.
        `Returns:`
            str
        """

        schema, view = self.split_full_table_name(view)

        if not self.is_view(f"{schema}.{view}"):
            return None

        results = self.get_view_definitions(schema, view)

        return results[0]["ddl"]

    def get_view_definitions(self, schema=None, view=None):
        """
        Get the view definition (i.e. the create statement) for multiple views.

        This works similar to `get_view_def` except it runs a single query
        to get the ddl for multiple views. It supports SQL wildcards for
        `schema` and `view`. Only returns the ddl for _views_ that match
        `schema` and `view` if they exist.

        `Args:`
            schema: str
                The schema to filter by.
            view: str
                The view to filter by.
        `Returns:`
            `list` of dicts with matching views.

        """

        conditions = []
        if schema:
            conditions.append(f"schemaname like '{schema}'")
        if view:
            conditions.append(f"g.viewname like '{view}'")

        conditions_str = ' and '.join(conditions)
        where_clause = f"where {conditions_str}" if conditions_str else ''

        sql_get_ddl = f"""
            select schemaname || '.' || viewname as viewname, ddl
            from admin.v_generate_view_ddl g
            {where_clause}
        """
        ddl_view = self.query(sql_get_ddl)

        if ddl_view.num_rows == 0:
            logger.info(f"No views matching {schema} and {view}.")
            return None

        return ddl_view.to_dicts()

    @staticmethod
    def split_full_table_name(full_table_name):
        """
        Split a full table name into its schema and table. If a schema isn't
        present, return `public` for the schema. Similarly, Redshift defaults
        to the `public` schema, when one isn't provided.

        Eg:
        ``(schema, table) = Redshift.split_full_table_name("some_schema.some_table")``

        `Args:`
            full_table_name: str
                The table name, as "schema.table"
        `Returns:`
            tuple
                A tuple containing (schema, table)
        """
        if "." not in full_table_name:
            return "public", full_table_name

        try:
            schema, table = full_table_name.split(".")
        except ValueError as e:
            if "too many values to unpack" in str(e):
                raise ValueError(f"Invalid Redshift table {full_table_name}")

        return schema, table

    @staticmethod
    def combine_schema_and_table_name(schema, table):
        """
        Creates a full table name by combining a schema and table.

        `Args:`
            schema: str
                The schema name
            table: str
                The table name
        `Returns:`
            str
                The combined full table name
        """
        return f"{schema}.{table}"



import google
from google.cloud import storage
from parsons.google.utitities import setup_google_application_credentials
from parsons.utilities import files
import logging

logger = logging.getLogger(__name__)


class GoogleCloudStorage(object):
    """
    This class requires application credentials in the form of a json. It can be passed
    in the following ways:

    * Set an environmental variable named ``GOOGLE_APPLICATION_CREDENTIALS`` with the
      local path to the credentials json.

      Example: ``GOOGLE_APPLICATION_CREDENTALS='path/to/creds.json'``

    * Pass in the path to the credentials using the ``app_creds`` argument.

    * Pass in a json string using the ``app_creds`` argument.


    `Args:`
        credentials: str
            A credentials json string or a path to a json file. Not required
            if ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set.
        project: str
            The project which the client is acting on behalf of. If not passed
            then will use the default inferred environment.
    `Returns:`
        GoogleCloudStorage Class
    """

    def __init__(self, app_creds=None, project=None):

        setup_google_application_credentials(app_creds)

        # Throws an error if you pass project=None, so adding if/else statement.
        if not project:
            self.client = storage.Client()
            """
            Access all methods of `google.cloud` package
            """
        else:
            self.client = storage.Client(project=project)

    def list_buckets(self):
        """
        Returns a list of buckets

        `Returns:`
            List of buckets
        """

        buckets = [b.name for b in self.client.list_buckets()]
        logger.info(f'Found {len(buckets)}.')
        return buckets

    def bucket_exists(self, bucket_name):
        """
        Verify that a bucket exists

        `Args:`
            bucket_name: str
                The name of the bucket
        `Returns:`
            boolean
        """

        if bucket_name in self.list_buckets():
            logger.info(f'{bucket_name} exists.')
            return True
        else:
            logger.info(f'{bucket_name} does not exist.')
            return False

    def get_bucket(self, bucket_name):
        """
        Returns a bucket object

        `Args:`
            bucket_name: str
                The name of bucket
        `Returns:`
            GoogleCloud Storage bucket
        """

        if self.client.lookup_bucket(bucket_name):
            bucket = self.client.get_bucket(bucket_name)
        else:
            raise google.cloud.exceptions.NotFound('Bucket not found')

        logger.info(f'Returning {bucket_name} object')
        return bucket

    def create_bucket(self, bucket_name):
        """
        Create a bucket.

        `Args:`
            bucket_name: str
                A globally unique name for the bucket.
        `Returns:`
            ``None``
        """

        # To Do: Allow user to set all of the bucket parameters

        self.client.create_bucket(bucket_name)
        logger.info(f'Created {bucket_name} bucket.')

    def delete_bucket(self, bucket_name, delete_blobs=False):
        """
        Delete a bucket. Will fail if not empty unless ``delete_blobs`` argument
        is set to ``True``.

        `Args:`
            bucket_name: str
                The name of the bucket
            delete_blobs: boolean
                Delete blobs in the bucket, if it is not empty
        `Returns:`
            ``None``
        """

        bucket = self.get_bucket(bucket_name)
        bucket.delete(force=delete_blobs)
        logger.info(f'{bucket_name} bucket deleted.')

    def list_blobs(self, bucket_name, max_results=None, prefix=None):
        """
        List all of the blobs in a bucket

        `Args:`
            bucket_name: str
                The name of the bucket
            max_results: int
                TBD
            prefix_filter: str
                A prefix to filter files
        `Returns:`
            A list of blob names
        """

        blobs = self.client.list_blobs(bucket_name, max_results=max_results, prefix=prefix)
        lst = [b.name for b in blobs]
        logger.info(f'Found {len(lst)} in {bucket_name} bucket.')

        return lst

    def blob_exists(self, bucket_name, blob_name):
        """
        Verify that a blob exists in the specified bucket

        `Args:`
            bucket_name: str
                The bucket name
            blob_name: str
                The name of the blob
        `Returns:`
            boolean
        """

        if blob_name in self.list_blobs(bucket_name):
            logger.info(f'{blob_name} exists.')
            return True
        else:
            logger.info(f'{blob_name} does not exist.')
            return False

    def get_blob(self, bucket_name, blob_name):
        """
        Get a blob object

        `Args:
            bucket_name: str
                A bucket name
            blob_name: str
                A blob name
        `Returns:`
            A Google Storage blob object
        """

        bucket = self.get_bucket(bucket_name)
        blob = bucket.get_blob(blob_name)
        logger.debug(f'Got {blob_name} object from {bucket_name} bucket.')
        return blob

    def put_blob(self, bucket_name, blob_name, local_path):
        """
        Puts a blob (aka file) in a bucket

        `Args:`
            blob_name:
                The name of blob to be stored in the bucket
            bucket_name:
                The name of the bucket to store the blob
            local_path: str
                The local path of the file to upload
        `Returns:`
            ``None``
        """

        bucket = self.get_bucket(bucket_name)
        blob = storage.Blob(blob_name, bucket)

        with open(local_path, "rb") as f:
            blob.upload_from_file(f)

        logger.info(f'{blob_name} put in {bucket_name} bucket.')

    def download_blob(self, bucket_name, blob_name, local_path=None):
        """
        Gets a blob from a bucket

        `Args:`
            bucket_name: str
                The name of the bucket
            blob_name: str
                The name of the blob
            local_path: str
                The local path where the file will be downloaded. If not specified, a temporary
                file will be created and returned, and that file will be removed automatically
                when the script is done running.
        `Returns:`
            str
                The path of the downloaded file
        """

        if not local_path:
            local_path = files.create_temp_file_for_path('TEMPTHING')

        bucket = storage.Bucket(self.client, name=bucket_name)
        blob = storage.Blob(blob_name, bucket)

        logger.info(f'Downloading {blob_name} from {bucket_name} bucket.')
        with open(local_path, 'wb') as f:
            blob.download_to_file(f, client=self.client)
        logger.info(f'{blob_name} saved to {local_path}.')

        return local_path

    def delete_blob(self, bucket_name, blob_name):
        """
        Delete a blob

        `Args:`
            bucket_name: str
                The bucket name
            blob_name: str
                The blob name
        `Returns:`
            ``None``
        """

        blob = self.get_blob(bucket_name, blob_name)
        blob.delete()
        logger.info(f'{blob_name} blob in {bucket_name} bucket deleted.')

    def upload_table(self, table, bucket_name, blob_name, data_type='csv', default_acl=None):
        """
        Load the data from a Parsons table into a blob.

        `Args:`
            table: obj
                A :ref:`parsons-table`
            bucket_name: str
                The name of the bucket to upload the data into.
            blob_name: str
                The name of the blob to upload the data into.
            data_type: str
                The file format to use when writing the data. One of: `csv` or `json`
        """
        bucket = storage.Bucket(self.client, name=bucket_name)
        blob = storage.Blob(blob_name, bucket)

        if data_type == 'csv':
            local_file = table.to_csv()
            content_type = 'text/csv'
        elif data_type == 'json':
            local_file = table.to_json()
            content_type = 'application/json'
        else:
            raise ValueError(f'Unknown data_type value ({data_type}): must be one of: csv or json')

        try:
            blob.upload_from_filename(local_file, content_type=content_type, client=self.client,
                                      predefined_acl=default_acl)
        finally:
            files.close_temp_file(local_file)

        return f'gs://{bucket_name}/{blob_name}'



from parsons import Postgres, DBSync, Table, Redshift
from test.utils import assert_matching_tables
import unittest
import os

_dir = os.path.dirname(__file__)

TEMP_SCHEMA = 'parsons_test'


# These tests interact directly with the Postgres database. In order to run, set the
# env to LIVE_TEST='TRUE'.
@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestPostgresDBSync(unittest.TestCase):

    def setUp(self):

        self.temp_schema = TEMP_SCHEMA
        self.db = Postgres()

        # Create a schema.
        setup_sql = f"""
                     DROP SCHEMA IF EXISTS {self.temp_schema} CASCADE;
                     CREATE SCHEMA {self.temp_schema};
                     """
        self.db.query(setup_sql)

        # Load dummy data to parsons tables
        self.table1 = Table.from_csv(f'{_dir}/test_data/sample_table_1.csv')
        self.table2 = Table.from_csv(f'{_dir}/test_data/sample_table_2.csv')

        # Create source table
        self.db.copy(self.table1, f'{self.temp_schema}.source')

        # Create DB Sync object
        self.db_sync = DBSync(self.db, self.db)

    def tearDown(self):
        # Drop the view, the table and the schema

        teardown_sql = f"""
                       DROP SCHEMA IF EXISTS {self.temp_schema} CASCADE;
                       """
        self.db.query(teardown_sql)

    def test_table_sync_full_drop(self):
        # Test a db sync with drop.

        self.db_sync.table_sync_full(f'{self.temp_schema}.source',
                                     f'{self.temp_schema}.destination')

        source = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        destination = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(source, destination)

    def test_table_sync_full_truncate(self):
        # Test a db sync with truncate.

        self.db_sync.table_sync_full(f'{self.temp_schema}.source',
                                      f'{self.temp_schema}.destination',
                                     if_exists='truncate')
        source = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        destination = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(source, destination)

    def test_table_sync_full_empty_table(self):
        # Test a full sync of a table when the source table is empty.

        # Empty the source table
        self.db.query(f"TRUNCATE {self.temp_schema}.source")

        # Attempt to sync.
        self.db_sync.table_sync_full(f'{self.temp_schema}.source',
                                      f'{self.temp_schema}.destination')


    def test_table_sync_full_chunk(self):
        # Test chunking in full sync.

        self.db_sync.chunk_size = 10
        self.db_sync.table_sync_full(f'{self.temp_schema}.source',
                                     f'{self.temp_schema}.destination')

        source = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        destination = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(source, destination)

    def test_table_sync_incremental(self):
        # Test that incremental sync

        self.db.copy(self.table1, f'{self.temp_schema}.destination')
        self.db.copy(self.table2, f'{self.temp_schema}.source', if_exists='append')
        self.db_sync.table_sync_incremental(f'{self.temp_schema}.source',
                                            f'{self.temp_schema}.destination',
                                            'pk')

        count1 = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        count2 = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(count1, count2)

    def test_table_sync_incremental_chunk(self):
        # Test chunking of incremental sync.

        self.db_sync.chunk_size = 10
        self.db.copy(self.table1, f'{self.temp_schema}.destination')
        self.db.copy(self.table2, f'{self.temp_schema}.source', if_exists='append')
        self.db_sync.table_sync_incremental(f'{self.temp_schema}.source',
                                            f'{self.temp_schema}.destination',
                                            'pk')

        count1 = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        count2 = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(count1, count2)

    def test_table_sync_incremental_create_destination_table(self):
        # Test that an incremental sync works if the destination table does not exist.

        self.db_sync.table_sync_incremental(f'{self.temp_schema}.source',
                                            f'{self.temp_schema}.destination',
                                            'pk')

        count1 = self.db.query(f"SELECT * FROM {self.temp_schema}.source")
        count2 = self.db.query(f"SELECT * FROM {self.temp_schema}.destination")
        assert_matching_tables(count1, count2)

    def test_table_sync_incremental_empty_table(self):
        # Test an incremental sync of a table when the source table is empty.

        # Empty the source table
        self.db.query(f"TRUNCATE {self.temp_schema}.source")

        # Attempt to sync.
        self.db_sync.table_sync_incremental(f'{self.temp_schema}.source',
                                            f'{self.temp_schema}.destination',
                                            'pk')

# These tests interact directly with the Postgres database. In order to run, set the
# env to LIVE_TEST='TRUE'.
@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestRedshiftDBSync(TestPostgresDBSync):
    """ This test inherits all of the tests from the Postgres test. """

    def setUp(self):

        self.temp_schema = TEMP_SCHEMA
        self.db = Redshift()

        # Create a schema.
        setup_sql = f"""
                     DROP SCHEMA IF EXISTS {self.temp_schema} CASCADE;
                     CREATE SCHEMA {self.temp_schema};
                     """
        self.db.query(setup_sql)

        # Load dummy data to parsons tables
        self.table1 = Table.from_csv(f'{_dir}/test_data/sample_table_1.csv')
        self.table2 = Table.from_csv(f'{_dir}/test_data/sample_table_2.csv')

        # Create source table
        self.db.copy(self.table1, f'{self.temp_schema}.source')

        # Create DB Sync object
        self.db_sync = DBSync(self.db, self.db)



"""
This utility method is a generalizable method for moving files to an
online file storage class. It is used by methods that require access
to a file via a public url (e.g. VAN). Currently only includes S3, but
in the future, might include SFTP, and Google Cloud Storage.
"""


def post_file(tbl, type, file_path=None, **file_storage_args):
    """
    This utility method is a generalizable method for moving files to an
    online file storage class. It is used by methods that require access
    to a file via a public url (e.g. VAN).

    **S3 is the only option allowed.**

    `Args:`
        tbl: object
            parsons.Table
        type: str
            ``S3``
        file_path: str
            The file path to store the file. Not required if provided with
            the **file_storage_args.
        **kwargs: kwargs
                Optional arguments specific to the file storage.
    `Returns:`
        ``None``
    """

    if type.capitalize() == 'S3':

        # Overwrite the file_path if key is passed
        if 'key' in file_storage_args:
            file_storage_args['key'] = file_path

        return tbl.to_s3_csv(public_url=True, key=file_path, **file_storage_args)

    # To Do: Add in SFTP, GoogleCloud Storage

    else:
        raise ValueError('Type must be S3.')



import os
import unittest
import requests_mock
from test.utils import assert_matching_tables

from parsons.etl.table import Table
from parsons.redash.redash import Redash, RedashTimeout

BASE_URL = 'https://redash.example.com'
API_KEY = 'abc123'


class TestRedash(unittest.TestCase):

    mock_data = 'foo,bar\n1,2\n3,4'
    mock_result = Table([('foo', 'bar'), ('1', '2'), ('3', '4')])

    def setUp(self):
        self.redash = Redash(BASE_URL, API_KEY)

    @requests_mock.Mocker()
    def test_cached_query(self, m):
        redash = Redash(BASE_URL)  # no user_api_key
        m.get(f'{BASE_URL}/api/queries/5/results.csv', text=self.mock_data)
        assert_matching_tables(redash.get_cached_query_results(5, API_KEY),
                               self.mock_result)
        self.assertEqual(m._adapter.last_request.path, '/api/queries/5/results.csv')
        self.assertEqual(m._adapter.last_request.query, 'api_key=abc123')

        assert_matching_tables(self.redash.get_cached_query_results(5),
                               self.mock_result)
        self.assertEqual(m._adapter.last_request.query, '')

    @requests_mock.Mocker()
    def test_refresh_query(self, m):
        m.post(f'{BASE_URL}/api/queries/5/refresh', json={
            'job': {'status': 3, 'query_result_id': 21}})
        m.get(f'{BASE_URL}/api/queries/5/results/21.csv', text=self.mock_data)

        assert_matching_tables(self.redash.get_fresh_query_results(5, {'yyy': 'xxx'}),
                               self.mock_result)

    @requests_mock.Mocker()
    def test_refresh_query_poll(self, m):
        m.post(f'{BASE_URL}/api/queries/5/refresh', json={
            'job': {'id': 66, 'status': 1}})
        m.get(f'{BASE_URL}/api/jobs/66', json={
            'job': {'id': 66, 'status': 3, 'query_result_id': 21}})
        m.get(f'{BASE_URL}/api/queries/5/results/21.csv', text=self.mock_data)

        self.redash.pause = 0.01  # shorten pause time
        assert_matching_tables(self.redash.get_fresh_query_results(5, {'yyy': 'xxx'}),
                               self.mock_result)

    @requests_mock.Mocker()
    def test_refresh_query_poll_timeout(self, m):
        m.post(f'{BASE_URL}/api/queries/5/refresh', json={
            'job': {'id': 66, 'status': 1}})
        m.get(f'{BASE_URL}/api/jobs/66', json={
            'job': {'id': 66, 'status': 1}})
        m.get(f'{BASE_URL}/api/queries/5/results/21.csv', text=self.mock_data)

        self.redash.pause = 0.01  # shorten pause time
        self.redash.timeout = 0.01  # timeout
        raised = False
        try:
            self.redash.get_fresh_query_results(5, {'yyy': 'xxx'})
        except RedashTimeout:
            raised = True
        self.assertTrue(raised)

    @requests_mock.Mocker()
    def test_to_table(self, m):
        m.post(f'{BASE_URL}/api/queries/5/refresh', json={
            'job': {'status': 3, 'query_result_id': 21}})
        m.get(f'{BASE_URL}/api/queries/5/results/21.csv', text=self.mock_data)

        self.redash.pause = 0.01  # shorten pause time
        table_data = Redash.load_to_table(
            base_url=BASE_URL,
            user_api_key=API_KEY,
            query_id=5,
            params={'x': 'y'},
            verify=False)

        assert_matching_tables(table_data, self.mock_result)

    @requests_mock.Mocker()
    def test_to_table_env_vars(self, m):
        try:
            _environ = dict(os.environ)
            os.environ.update({
                'REDASH_BASE_URL': BASE_URL,
                'REDASH_USER_API_KEY': API_KEY,
                'REDASH_QUERY_ID': "5",
                'REDASH_QUERY_PARAMS': "p_x=y"
            })
            m.post(f'{BASE_URL}/api/queries/5/refresh', json={
                'job': {'status': 3, 'query_result_id': 21}})
            m.get(f'{BASE_URL}/api/queries/5/results/21.csv', text=self.mock_data)

            self.redash.pause = 0.01  # shorten pause time

            assert_matching_tables(Redash.load_to_table(), self.mock_result)
        finally:
            os.environ.clear()
            os.environ.update(_environ)



import logging

import braintree

from parsons.etl.table import Table
from parsons.utilities.check_env import check as check_env

logger = logging.getLogger(__name__)


class ParsonsBraintreeError(Exception):
    pass


class Braintree(object):
    """
    Braintree is a payment processor.
    `Args:`
        merchant_id: str
            Braintree merchant id -- probably a 16-char alphanumeric.
            Not required if ``BRAINTREE_MERCHANT_ID`` env variable is set.
        public_key: str
            Braintree public key -- probably a (different) 16-char alphanumeric.
            Not required if ``BRAINTREE_PUBLIC_KEY`` env variable is set.
        private_key: str
            Braintree private key -- probably a 32-char alphanumeric.
            Not required if ``BRAINTREE_PRIVATE_KEY`` env variable is set.
        timeout: int
            Optionally change the timeout from default of 200 seconds.
            Can also be passed with env var ``BRAINTREE_TIMEOUT``.
        production: bool
            Defaults to True.  If you are testing in a Sandbox,
            set this to False.
    `Returns:`
        Braintree class
    """

    query_types = {
        'dispute': braintree.DisputeSearch,
        'transaction': braintree.TransactionSearch,
    }

    credit_card_fields = [
        'bin',
        'card_type',
        'cardholder_name',
        'commercial',
        'country_of_issuance',
        'customer_location',
        'debit',
        'durbin_regulated',
        'expiration_month',
        'expiration_year',
        'healthcare',
        'image_url',
        'issuing_bank',
        'last_4',
        'payroll',
        'prepaid',
        'product_id',
        'token',
        'venmo_sdk',
    ]

    disbursement_fields = [
        'disbursement_date',  # => disbursement_date column
        'funds_held',
        'settlement_amount',
        'settlement_currency_exchange_rate',
        'settlement_currency_iso_code',
        'success',
    ]

    transaction_fields = [
        'additional_processor_response',
        'amount',
        'avs_error_response_code',
        'avs_postal_code_response_code',
        'avs_street_address_response_code',
        'channel',
        'created_at',
        'currency_iso_code',
        'cvv_response_code',
        'discount_amount',
        'escrow_status',
        'gateway_rejection_reason',
        'id',
        'master_merchant_account_id',
        'merchant_account_id',
        'order_id',
        'payment_instrument_type',
        'plan_id',
        'processor_authorization_code',
        'processor_response_code',
        'processor_response_text',
        'processor_settlement_response_code',
        'processor_settlement_response_text',
        'purchase_order_number',
        'recurring',
        'refund_id',
        'refunded_transaction_id',
        'service_fee_amount',
        'settlement_batch_id',
        'shipping_amount',
        'ships_from_postal_code',
        'status',
        'sub_merchant_account_id',
        'subscription_id',
        'tax_amount',
        'tax_exempt',
        'type',
        'updated_at',
        'voice_referral_number',
    ]

    dispute_fields = [
        'id',
        'amount_disputed',
        'amount_won',
        'case_number',
        'currency_iso_code',
        'kind',
        'merchant_account_id',
        'original_dispute_id',
        'processor_comments',
        'reason',
        'reason_code',
        'reason_description',
        'received_date',
        'reference_number',
        'reply_by_date',
        'status',
        # 'transaction.id', # DOT id -- needs to be special-cased (below)
    ]

    def __init__(self, merchant_id=None, public_key=None, private_key=None,
                 timeout=None, production=True):
        merchant_id = check_env('BRAINTREE_MERCHANT_ID', merchant_id)
        public_key = check_env('BRAINTREE_PUBLIC_KEY', public_key)
        private_key = check_env('BRAINTREE_PRIVATE_KEY', private_key)
        timeout = check_env('BRAINTREE_TIMEOUT', timeout, optional=True) or 200

        self.gateway = braintree.BraintreeGateway(
            braintree.Configuration(
                environment=(braintree.Environment.Production if production
                             else braintree.Environment.Sandbox),
                merchant_id=merchant_id,
                public_key=public_key,
                private_key=private_key,
                timeout=timeout))

    def get_disputes(self,
                     start_date=None, end_date=None,
                     query_list=None,
                     query_dict=None):
        """
        Get a table of disputes based on query parameters.
        There are three ways to pass query arguments: Pass a start_date and end_date
        together for a date range, or pass a query_list or query_dict argument.

        `Args:`
            start_date: date or str
                Start date of the dispute range. Requires `end_date` arg. e.g. '2020-11-03'
            end_date: date or str
                End date of the dispute range. Requires `start_date` arg. e.g. '2020-11-03'
            query_list: list of braintree.DisputeSearch
                You can use the `braintree.DisputeSearch
                <https://developers.braintreepayments.com/reference/request/dispute/search/python>`_
                to create a manual list of query parameters.
            query_dict: jsonable-dict
                query_dict is basically the same as query_list, except instead of using their API
                objects, you can pass it in pure dictionary form.
                Some examples:
                    .. highlight:: python
                    .. code-block:: python

                      # The start_date/end_date arguments are the same as
                      {"effective_date": {"between": [start_date, end_date]}}
                      # some other examples
                      {"merchant_account_id": {"in_list": [123, 456]}}
                      {"created_at": {"greater_than_or_equal": "2020-03-10"}}

        `Returns:`
            Table Class
        """
        collection = self._get_collection(
            'dispute',
            query_list=query_list,
            query_dict=query_dict,
            default_query=(
                {'effective_date': dict(between=[start_date, end_date])}
                if start_date and end_date
                else None
            ))

        # Iterating on collection.items triggers web requests in batches of 50 records
        # Disputes query api doesn't return the ids -- we can't do anything but iterate
        if not collection.is_success:
            raise ParsonsBraintreeError(f"Braintree dispute query failed: {collection.message}")
        return Table([
            self._dispute_header()
        ] + [self._dispute_to_row(r) for r in collection.disputes.items])

    def get_transactions(self,
                         table_of_ids=None,
                         disbursement_start_date=None, disbursement_end_date=None,
                         query_list=None,
                         query_dict=None,
                         just_ids=False):
        """
        Get a table of transactions based on query parameters.
        There are three ways to pass query arguments:
        Pass a disbursement_start_date and disbursement_end_date together
        for a date range, or pass a query_list or query_dict argument.

        `Args:`
            disbursement_start_date: date or str
                Start date of the disbursement range. Requires `disbursement_end_date` arg.
                e.g. '2020-11-03'
            disbursement_end_date: date or str
                End date of the disbursement range. Requires `disbursement_start_date` arg.
                e.g. '2020-11-03'
            query_list: list of braintree.TransactionSearch
                You can use the `braintree.TransactionSearch
                <https://developers.braintreepayments.com/reference/request/transaction/search/python>`_
                to create a manual list of query parameters.
            query_dict: jsonable-dict
                query_dict is basically the same as query_list, except instead of using their API
                objects, you can pass it in pure dictionary form.
                Some examples:
                    .. highlight:: python
                    .. code-block:: python

                      # The disbursement_start_date/disbursement_end_date arguments are the same as
                      {"disbursement_date": {"between": [start_date, end_date]}}
                      # some other examples
                      {"merchant_account_id": {"in_list": [123, 456]}}
                      {"created_at": {"greater_than_or_equal": "2020-03-10"}}
            just_ids: bool
                While querying a list of transaction ids is a single, fast query to Braintree's API,
                getting all data for each transaction is force-paginated at 50-records per request.
                If you just need a count or the list of ids, then set `just_ids=True` and
                it will return a single column with `id` instead of all table columns.
            table_of_ids: Table with an `id` column -- i.e. a table returned from `just_ids=True`
                Subsequently, after calling this with `just_ids`, you can prune/alter the ids table
                and then pass the table back to get the full data.
                These are somewhat-niche use-cases, but occasionally crucial
                when a search result returns 1000s of ids.
        `Returns:`
            Table Class
        """
        collection = self._get_collection(
            'transaction',
            table_of_ids=table_of_ids,
            query_list=query_list,
            query_dict=query_dict,
            default_query=(
                {'disbursement_date': dict(
                    between=[disbursement_start_date, disbursement_end_date])}
                if disbursement_start_date and disbursement_end_date
                else None
            ))
        query_count = len(collection.ids)
        logger.info(f'Braintree transactions resulted in transaction count of {query_count}')
        if just_ids:
            return Table([('id',)]
                         + [[item_id] for item_id in collection.ids])

        # Iterating on collection.items triggers web requests in batches of 50 records
        # This can be frustratingly slow :-(
        # Also note: Braintree will push you to their new GraphQL API,
        #   but it, too, paginates with a max of 50 records
        logger.debug('Braintree transactions iterating to build transaction table')
        return Table([
            self._transaction_header()
        ] + [self._transaction_to_row(r) for r in collection.items])

    def _dispute_header(self):
        return self.dispute_fields + ['transaction_id']

    def _dispute_to_row(self, collection_item):
        row = [getattr(collection_item, k) for k in self.dispute_fields]
        # the single sub-attribute
        row.append(collection_item.transaction.id)
        return row

    def _transaction_header(self):
        return (
            [f'credit_card_{k}' for k in self.credit_card_fields]
            # annoying exception in column name
            + [(f'disbursement_{k}' if k != 'disbursement_date' else k)
               for k in self.disbursement_fields]
            + self.transaction_fields)

    def _transaction_to_row(self, collection_item):
        return (
            [(collection_item.credit_card.get(k)
              if getattr(collection_item, 'credit_card', None) else None)
             for k in self.credit_card_fields]
            + [getattr(collection_item.disbursement_details, k) for k in self.disbursement_fields]
            + [getattr(collection_item, k) for k in self.transaction_fields])

    def _get_collection(self, query_type,
                        table_of_ids=None,
                        query_list=None,
                        query_dict=None,
                        default_query=None):
        collection_query = None
        collection = None
        if query_list:
            collection_query = query_list
        elif query_dict:
            collection_query = self._get_query_objects(query_type, **query_dict)
        elif default_query:
            collection_query = self._get_query_objects(query_type, **default_query)

        if not collection_query:
            raise ParsonsBraintreeError(
                "You must pass some query parameters: "
                "query_dict, start_date with end_date, or query_list")

        if table_of_ids:
            # We don't need to re-do the query, we can just reconstruct the query object
            collection = self._create_collection(table_of_ids.table.values('id'), collection_query)
        else:
            collection = getattr(self.gateway, query_type).search(*collection_query)
        return collection

    def _get_query_objects(self, query_type, **queryparams):
        """
        Examples:
        disbursement_date={'between': ['2020-03-20', '2020-03-27']}
        merchant_account_id={'in_list': [123, 456]}
        created_at={'greater_than_or_equal': '2020-03-10'}
        """
        queries = []
        for node, filters in queryparams.items():
            # Very meta programming here, abstracting braintree library's (dumb?) attribute style
            # Example: braintree.DisputeSearch.effective_date
            # Example: braintree.TransactionSearch.disbursement_date
            queryobj = getattr(self.query_types[query_type], node, None)
            if queryobj:
                for qual, vals in filters.items():  # likely only one, but fine
                    queryobj_qualfunc = getattr(queryobj, qual, None)
                    if not queryobj_qualfunc:
                        raise ParsonsBraintreeError("oh no, that's not a braintree parameter")
                    if not isinstance(vals, list):
                        vals = [vals]
                    queries.append(queryobj_qualfunc(*vals))
            else:
                raise ParsonsBraintreeError("oh no, that's not a braintree parameter")
        return queries

    def _create_collection(self, ids, queries):
        transaction_gateway = braintree.TransactionGateway(self.gateway)
        return braintree.ResourceCollection(
            queries,
            {'search_results': {'ids': list(ids), 'page_size': 50}},
            method=transaction_gateway._TransactionGateway__fetch)



import pickle
import uuid

from google.cloud import bigquery
from google.cloud.bigquery import dbapi
from google.cloud import exceptions
import petl

from parsons.databases.table import BaseTable
from parsons.etl import Table
from parsons.google.utitities import setup_google_application_credentials
from parsons.google.google_cloud_storage import GoogleCloudStorage
from parsons.utilities import check_env
from parsons.utilities.files import create_temp_file

BIGQUERY_TYPE_MAP = {
    'str': 'STRING',
    'float': 'FLOAT',
    'int': 'INTEGER',
    'bool': 'BOOLEAN',
    'datetime.datetime': 'DATETIME',
    'datetime.date': 'DATE',
    'datetime.time': 'TIME',
    'dict': 'RECORD',
}

# Max number of rows that we query at a time, so we can avoid loading huge
# data sets into memory.
# 100k rows per batch at ~1k bytes each = ~100MB per batch.
QUERY_BATCH_SIZE = 100000


def get_table_ref(client, table_name):
    # Helper function to build a TableReference for our table
    parsed = parse_table_name(table_name)
    dataset_ref = client.dataset(parsed['dataset'])
    return dataset_ref.table(parsed['table'])


def parse_table_name(table_name):
    # Helper function to parse out the different components of a table ID
    parts = table_name.split('.')
    parts.reverse()
    parsed = {
        'project': None,
        'dataset': None,
        'table': None,
    }
    if len(parts) > 0:
        parsed['table'] = parts[0]
    if len(parts) > 1:
        parsed['dataset'] = parts[1]
    if len(parts) > 3:
        parsed['project'] = parts[2]
    return parsed


class GoogleBigQuery:
    """
    Class for querying BigQuery table and returning the data as Parsons tables.

    This class requires application credentials in the form of a json. It can be passed
    in the following ways:

    * Set an environmental variable named ``GOOGLE_APPLICATION_CREDENTIALS`` with the
      local path to the credentials json.

      Example: ``GOOGLE_APPLICATION_CREDENTALS='path/to/creds.json'``

    * Pass in the path to the credentials using the ``app_creds`` argument.

    * Pass in a json string using the ``app_creds`` argument.

    Args:
        project_id: str
            The project which the client is acting on behalf of. If not passed
            then will use the default inferred environment.
        app_creds: str
            A credentials json string or a path to a json file. Not required
            if ``GOOGLE_APPLICATION_CREDENTIALS`` env variable set.
        location: str
            Default geographic location for tables
    """

    def __init__(self, app_creds=None, project=None, location=None):
        self.app_creds = app_creds

        setup_google_application_credentials(app_creds)

        self.project = project
        self.location = location

        # We will not create the client until we need to use it, since creating the client
        # without valid GOOGLE_APPLICATION_CREDENTIALS raises an exception.
        # This attribute will be used to hold the client once we have created it.
        self._client = None

        self._dbapi = dbapi

        self.dialect = 'bigquery'

    def copy(self, table_obj, table_name, if_exists='fail',
             tmp_gcs_bucket=None, gcs_client=None, job_config=None, **load_kwargs):
        """
        Copy a :ref:`parsons-table` into Google BigQuery via Google Cloud Storage.

        `Args:`
            table_obj: obj
                The Parsons Table to copy into BigQuery.
            table_name: str
                The table name to load the data into.
            if_exists: str
                If the table already exists, either ``fail``, ``append``, ``drop``
                or ``truncate`` the table.
            temp_gcs_bucket: str
                The name of the Google Cloud Storage bucket to use to stage the data to load
                into BigQuery. Required if `GCS_TEMP_BUCKET` is not specified.
            gcs_client: object
                The GoogleCloudStorage Connector to use for loading data into Google Cloud Storage.
            job_config: object
                A LoadJobConfig object to provide to the underlying call to load_table_from_uri
                on the BigQuery client. The function will create its own if not provided.
            **load_kwargs: kwargs
                Arguments to pass to the underlying load_table_from_uri call on the BigQuery
                client.
        """
        tmp_gcs_bucket = check_env.check('GCS_TEMP_BUCKET', tmp_gcs_bucket)

        if if_exists not in ['fail', 'truncate', 'append', 'drop']:
            raise ValueError(f'Unexpected value for if_exists: {if_exists}, must be one of '
                             '"append", "drop", "truncate", or "fail"')

        table_exists = self.table_exists(table_name)

        if not job_config:
            job_config = bigquery.LoadJobConfig()

        if not job_config.schema:
            job_config.schema = self._generate_schema(table_obj)

        if not job_config.create_disposition:
            job_config.create_disposition = bigquery.CreateDisposition.CREATE_IF_NEEDED
        job_config.skip_leading_rows = 1
        job_config.source_format = bigquery.SourceFormat.CSV
        job_config.write_disposition = bigquery.WriteDisposition.WRITE_EMPTY

        if table_exists:
            if if_exists == 'fail':
                raise ValueError('Table already exists.')
            elif if_exists == 'drop':
                self.delete_table(table_name)
            elif if_exists == 'append':
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
            elif if_exists == 'truncate':
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE

        gcs_client = gcs_client or GoogleCloudStorage()
        temp_blob_name = f'{uuid.uuid4()}.csv'
        temp_blob_uri = gcs_client.upload_table(table_obj, tmp_gcs_bucket, temp_blob_name)

        # load CSV from Cloud Storage into BigQuery
        table_ref = get_table_ref(self.client, table_name)
        try:
            load_job = self.client.load_table_from_uri(
                temp_blob_uri, table_ref,
                job_config=job_config, **load_kwargs,
            )
            load_job.result()
        finally:
            gcs_client.delete_blob(tmp_gcs_bucket, temp_blob_name)

    def delete_table(self, table_name):
        """
        Delete a BigQuery table.

        `Args:`
            table_name: str
                The name of the table to delete.
        """
        table_ref = get_table_ref(self.client, table_name)
        self.client.delete_table(table_ref)

    def query(self, sql, parameters=None):
        """
        Run a BigQuery query and return the results as a Parsons table.

        To include python variables in your query, it is recommended to pass them as parameters,
        following the BigQuery style where parameters are prefixed with `@`s.
        Using the ``parameters`` argument ensures that values are escaped properly, and avoids SQL
        injection attacks.

        **Parameter Examples**

        .. code-block:: python

        name = "Beatrice O'Brady"
        sql = 'SELECT * FROM my_table WHERE name = %s'
        rs.query(sql, parameters=[name])

        .. code-block:: python

        name = "Beatrice O'Brady"
        sql = "SELECT * FROM my_table WHERE name = %(name)s"
        rs.query(sql, parameters={'name': name})

        `Args:`
            sql: str
                A valid BigTable statement
            parameters: dict
                A dictionary of query parameters for BigQuery.

        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """
        # get our connection and cursor
        cursor = self._dbapi.connect(self.client).cursor()

        # Run the query
        cursor.execute(sql, parameters)

        # We will use a temp file to cache the results so that they are not all living
        # in memory. We'll use pickle to serialize the results to file in order to maintain
        # the proper data types (e.g. integer).
        temp_filename = create_temp_file()

        wrote_header = False
        with open(temp_filename, 'wb') as temp_file:
            # Track whether we got data, since if we don't get any results we need to return None
            got_results = False
            while True:
                batch = cursor.fetchmany(QUERY_BATCH_SIZE)
                if len(batch) == 0:
                    break

                got_results = True

                for row in batch:
                    # Make sure we write out the header once and only once
                    if not wrote_header:
                        wrote_header = True
                        header = list(row.keys())
                        pickle.dump(header, temp_file)

                    row_data = list(row.values())
                    pickle.dump(row_data, temp_file)

        if not got_results:
            return None

        ptable = petl.frompickle(temp_filename)
        final_table = Table(ptable)

        return final_table

    def table_exists(self, table_name):
        """
        Check whether or not the Google BigQuery table exists in the specified dataset.

        `Args:`
            table_name: str
                The name of the BigQuery table to check for
        `Returns:`
            bool
                True if the table exists in the specified dataset, false otherwise
        """
        table_ref = get_table_ref(self.client, table_name)
        try:
            self.client.get_table(table_ref)
        except exceptions.NotFound:
            return False

        return True

    @property
    def client(self):
        """
        Get the Google BigQuery client to use for making queries.

        `Returns:`
            `google.cloud.bigquery.client.Client`
        """
        if not self._client:
            # Create a BigQuery client to use to make the query
            self._client = bigquery.Client(project=self.project, location=self.location)

        return self._client

    def _generate_schema(self, tbl):
        stats = tbl.get_columns_type_stats()
        fields = []
        for stat in stats:
            petl_types = stat['type']
            best_type = 'str' if 'str' in petl_types else petl_types[0]
            field_type = self._bigquery_type(best_type)
            field = bigquery.schema.SchemaField(stat['name'], field_type)
            fields.append(field)
        return fields

    @staticmethod
    def _bigquery_type(tp):
        return BIGQUERY_TYPE_MAP[tp]

    def table(self, table_name):
        # Return a MySQL table object

        return BigQueryTable(self, table_name)


class BigQueryTable(BaseTable):
    # BigQuery table object.

    def drop(self, cascade=False):
        """
        Drop the table.
        """

        self.db.delete_table(self.table)

    def truncate(self):
        """
        Truncate the table.
        """
        # BigQuery does not support truncate natively, so we will "load" an empty dataset
        # with write disposition of "truncate"
        table_ref = get_table_ref(self.db.client, self.table)
        bq_table = self.db.client.get_table(table_ref)

        # BigQuery wants the schema when we load the data, so we will grab it from the table
        job_config = bigquery.LoadJobConfig()
        job_config.schema = bq_table.schema

        empty_table = Table([])
        self.db.copy(empty_table, self.table, if_exists='truncate', job_config=job_config)



from parsons.databases.postgres import Postgres
from parsons.etl.table import Table
from test.utils import assert_matching_tables
import unittest
import os
import re
import warnings
import datetime
import shutil
from test.utils import validate_list

# The name of the schema and will be temporarily created for the tests
TEMP_SCHEMA = 'parsons_test'

# These tests do not interact with the Postgres Database directly, and don't need real credentials


class TestPostgresCreateStatement(unittest.TestCase):

    def setUp(self):

        self.pg = Postgres(username='test', password='test', host='test', db='test', port=123)

        self.tbl = Table([['ID', 'Name'],
                          [1, 'Jim'],
                          [2, 'John'],
                          [3, 'Sarah']])

        self.mapping = self.pg.generate_data_types(self.tbl)

    def test_connection(self):

        # Test connection with kwargs passed
        arg_pg = Postgres(username='test', password='test', host='test', db='test')

        # Test connection with env variables
        os.environ['PGUSER'] = 'user_env'
        os.environ['PGPASSWORD'] = 'pass_env'
        os.environ['PGHOST'] = 'host_env'
        os.environ['PGDATABASE'] = 'db_env'
        os.environ['PGPORT'] = '5432'
        pg_env = Postgres()

        self.assertEqual(pg_env.username, 'user_env')
        self.assertEqual(pg_env.password, 'pass_env')
        self.assertEqual(pg_env.host, 'host_env')
        self.assertEqual(pg_env.db, 'db_env')
        self.assertEqual(pg_env.port, 5432)

    def test_data_type(self):

        # Test smallint
        self.assertEqual(self.pg.data_type(1, ''), 'smallint')
        # Test int
        self.assertEqual(self.pg.data_type(32769, ''), 'int')
        # Test bigint
        self.assertEqual(self.pg.data_type(2147483648, ''), 'bigint')
        # Test varchar that looks like an int
        self.assertEqual(self.pg.data_type('00001', ''), 'varchar')
        # Test a float as a decimal
        self.assertEqual(self.pg.data_type(5.001, ''), 'decimal')
        # Test varchar
        self.assertEqual(self.pg.data_type('word', ''), 'varchar')
        # Test int with underscore
        self.assertEqual(self.pg.data_type('1_2', ''), 'varchar')
        # Test int with leading zero
        self.assertEqual(self.pg.data_type('01', ''), 'varchar')

    def test_generate_data_types(self):

        # Test correct header labels
        self.assertEqual(self.mapping['headers'], ['ID', 'Name'])
        # Test correct data types
        self.assertEqual(self.mapping['type_list'], ['smallint', 'varchar'])
        # Test correct lengths
        self.assertEqual(self.mapping['longest'], [1, 5])

    def test_vc_padding(self):

        # Test padding calculated correctly
        self.assertEqual(self.pg.vc_padding(self.mapping, .2), [1, 6])

    def test_vc_max(self):

        # Test max sets it to the max
        self.assertEqual(self.pg.vc_max(self.mapping, ['Name']), [1, 65535])

        # Test raises when can't find column
        # To Do

    def test_vc_validate(self):

        # Test that a column with a width of 0 is set to 1
        self.mapping['longest'][0] = 0
        self.mapping = self.pg.vc_validate(self.mapping)
        self.assertEqual(self.mapping, [1, 5])

    def test_create_sql(self):

        # Test the the statement is expected
        sql = self.pg.create_sql('tmc.test', self.mapping, distkey='ID')
        exp_sql = "create table tmc.test (\n  id smallint,\n  name varchar(5)) \ndistkey(ID) ;"
        self.assertEqual(sql, exp_sql)

    def test_column_validate(self):

        bad_cols = ['', 'SELECT', 'asdfjkasjdfklasjdfklajskdfljaskldfjaklsdfjlaksdfjklasjdfklasjdkfljaskldfljkasjdkfasjlkdfjklasdfjklakjsfasjkdfljaslkdfjklasdfjklasjkldfakljsdfjalsdkfjklasjdfklasjdfklasdkljf'] # noqa: E501
        fixed_cols = ['col_0', 'col_1', 'asdfjkasjdfklasjdfklajskdfljaskldfjaklsdfjlaksdfjklasjdfklasjdkfljaskldfljkasjdkfasjlkdfjklasdfjklakjsfasjkdfljaslkdfjkl'] # noqa: E501
        self.assertEqual(self.pg.column_name_validate(bad_cols), fixed_cols)

    def test_create_statement(self):

        # Assert that copy statement is expected
        sql = self.pg.create_statement(self.tbl, 'tmc.test', distkey='ID')
        exp_sql = """create table tmc.test (\n  "id" smallint,\n  "name" varchar(5)) \ndistkey(ID) ;"""  # noqa: E501
        self.assertEqual(sql, exp_sql)

        # Assert that an error is raised by an empty table
        empty_table = Table([['Col_1', 'Col_2']])
        self.assertRaises(ValueError, self.pg.create_statement, empty_table, 'tmc.test')

# These tests interact directly with the Postgres database

@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestPostgresDB(unittest.TestCase):

    def setUp(self):

        self.temp_schema = TEMP_SCHEMA
        self.pg = Postgres()

        self.tbl = Table([['ID', 'Name'],
                          [1, 'Jim'],
                          [2, 'John'],
                          [3, 'Sarah']])

        # Create a schema, create a table, create a view
        setup_sql = f"""
                    drop schema if exists {self.temp_schema} cascade;
                    create schema {self.temp_schema};
                    """

        other_sql = f"""
                    create table {self.temp_schema}.test (id smallint,name varchar(5));
                    create view {self.temp_schema}.test_view as (select * from {self.temp_schema}.test);
                    """ # noqa: E501

        self.pg.query(setup_sql)

        self.pg.query(other_sql)

    def tearDown(self):

        # Drop the view, the table and the schema
        teardown_sql = f"""
                       drop schema if exists {self.temp_schema} cascade;
                       """
        self.pg.query(teardown_sql)

    def test_query(self):

        # Check that query sending back expected result
        r = self.pg.query('select 1')
        self.assertEqual(r[0]['?column?'], 1)

    def test_query_with_parameters(self):
        table_name = f"{self.temp_schema}.test"
        self.pg.copy(self.tbl, f"{self.temp_schema}.test", if_exists='append')

        sql = f"select * from {table_name} where name = %s"
        name = 'Sarah'
        r = self.pg.query(sql, parameters=[name])
        self.assertEqual(r[0]['name'], name)

        sql = f"select * from {table_name} where name in (%s, %s)"
        names = ['Sarah', 'John']
        r = self.pg.query(sql, parameters=names)
        self.assertEqual(r.num_rows, 2)

    def test_copy(self):

        # Copy a table and ensure table exists
        self.pg.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')
        r = self.pg.query(f"select * from {self.temp_schema}.test_copy where name='Jim'")
        self.assertEqual(r[0]['id'], 1)

        # Copy table and ensure truncate works.
        self.pg.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='truncate')
        tbl = self.pg.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(tbl.first, 3)

        # Copy table and ensure that drop works.
        self.pg.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')
        tbl = self.pg.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(tbl.first, 3)

        # Copy table and ensure that append works.
        self.pg.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='append')
        tbl = self.pg.query(f"select count(*) from {self.temp_schema}.test_copy")
        self.assertEqual(tbl.first, 6)

        # Try to copy the table and ensure that default fail works.
        self.assertRaises(ValueError, self.pg.copy, self.tbl, f'{self.temp_schema}.test_copy')

        # Try to copy the table and ensure that explicit fail works.
        self.assertRaises(ValueError, self.pg.copy, self.tbl, f'{self.temp_schema}.test_copy', if_exists='fail')

    def test_to_postgres(self):

        self.tbl.to_postgres(f'{self.temp_schema}.test_copy')
        r = self.pg.query(f"select * from {self.temp_schema}.test_copy where name='Jim'")
        self.assertEqual(r[0]['id'], 1)

    def test_from_postgres(self):

        tbl = Table([['id', 'name'],
                     [1, 'Jim'],
                     [2, 'John'],
                     [3, 'Sarah']])

        self.pg.copy(self.tbl, f'{self.temp_schema}.test_copy', if_exists='drop')
        out_tbl = self.tbl.from_postgres(f"SELECT * FROM {self.temp_schema}.test_copy")
        assert_matching_tables(out_tbl, tbl)

if __name__ == "__main__":
    unittest.main()




test_agent = [{
  'available': False,
  'occasional': True,
  'id': 47020956237,
  'signature': '<div dir="ltr"><p><br></p>\n</div>',
  'ticket_scope': 1,
  'created_at': '2020-01-17T15:07:01Z',
  'updated_at': '2020-02-05T00:58:37Z',
  'available_since': None,
  'type': 'support_agent',
  'contact': {
    'active': True,
    'email': 'person@email.org',
    'job_title': None,
    'language': 'en',
    'last_login_at': '2020-01-24T22:49:52Z',
    'mobile': None,
    'name': 'Alissa',
    'phone': None,
    'time_zone': 'Bogota',
    'created_at': '2020-01-17T15:07:00Z',
    'updated_at': '2020-01-24T22:49:52Z'
  }
}]

test_ticket = [{
  'cc_emails': ['person@email.org', 'person2@email.org'],
  'fwd_emails': [],
  'reply_cc_emails': ['person@email.org', 'person2@email.org'],
  'ticket_cc_emails': ['person@email.org', 'person2@email.org'],
  'fr_escalated': False,
  'spam': False,
  'email_config_id': None,
  'group_id': 47000643034,
  'priority': 1,
  'requester_id': 47021937449,
  'responder_id': 47017224681,
  'source': 3,
  'company_id': 47000491688,
  'status': 5,
  'subject': 'My thing is broken.',
  'association_type': None,
  'to_emails': None,
  'product_id': None,
  'id': 84,
  'type': 'Support Request 1',
  'due_by': '2020-02-19T22:00:00Z',
  'fr_due_by': '2020-02-06T17:00:00Z',
  'is_escalated': False,
  'custom_fields': {},
  'created_at': '2020-02-05T22:17:41Z',
  'updated_at': '2020-02-06T02:07:37Z',
  'associated_tickets_count': None,
  'tags': []
}]

test_company = [{
  'id': 47000491701,
  'name': 'Big Org',
  'description': None,
  'note': None,
  'domains': [],
  'created_at': '2020-01-09T20:43:09Z',
  'updated_at': '2020-01-09T20:43:09Z',
  'custom_fields': {},
  'health_score': None,
  'account_tier': 'Tier 2',
  'renewal_date': '2020-12-31T00:00:00Z',
  'industry': None
}]

test_contact = [{
  'active': False,
  'address': None,
  'company_id': 47000491686,
  'description': None,
  'email': 'person@email.org',
  'id': 47021299020,
  'job_title': None,
  'language': 'en',
  'mobile': None,
  'name': 'Percy Person',
  'phone': 'N/A',
  'time_zone': 'Bogota',
  'twitter_id': None,
  'custom_fields': {},
  'facebook_id': None,
  'created_at': '2020-01-27T16:44:34Z',
  'updated_at': '2020-01-27T16:44:34Z',
  'unique_external_id': None
}]




"""NGPVAN Code Endpoints"""
from parsons.etl.table import Table
import logging

logger = logging.getLogger(__name__)


class Codes(object):

    def __init__(self, van_connection):

        self.connection = van_connection

    def get_codes(self, name=None, supported_entities=None, parent_code_id=None,
                  code_type=None):
        """
        Get codes.

        `Args:`
            name : str
                Filter by name of code.
            supported_entities: str
                Filter by supported entities.
            parent_code_id: str
                Filter by parent code id.
            code_type: str
                Filter by code type.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        params = {'name': name,
                  'supportedEntities': supported_entities,
                  'parentCodeId': parent_code_id,
                  'codeType': code_type,
                  '$top': 200
                  }

        tbl = Table(self.connection.get_request('codes', params=params))
        logger.info(f'Found {tbl.num_rows} codes.')
        return tbl

    def get_code(self, code_id):
        """
        Get a code.

        `Args:`
            code_id : int
                The code id.
        `Returns:`
            Parsons Table
                See :ref:`parsons-table` for output options.
        """

        c = self.connection.request(f'codes/{code_id}')
        logger.debug(c)
        logger.info(f'Found code {code_id}.')
        return c

    def get_code_types(self):
        """
        Get code types.

        `Returns:`
            list
                A list of code types.
        """

        lst = self.connection.get_request('codeTypes')
        logger.info(f'Found {len(lst)} code types.')
        return lst

    def create_code(self, name=None, parent_code_id=None, description=None,
                    code_type='SourceCode', supported_entities=None):
        """
        Create a code.

        `Args:`
            name: str
                The name of the code.
            parent_code_id: int
                A unique identifier for this code’s parent.
            description: str
                A description for this code, no longer than 200 characters.
            code_type: str
                The code type. ``Tag`` and ``SourceCode`` are valid values.
            supported_entities: list
                A list of dicts that enumerate the searchability and applicability rules of the
                code. You can find supported entities with the :meth:`code_supported_entities`

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Event',
                         'is_searchable': True,
                         'is_applicable': True
                        }
                        {
                         'name': 'Locations',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]
        """

        json = {"parentCodeId": parent_code_id,
                "name": name,
                "codeType": code_type,
                "description": description}

        if supported_entities:

            se = [{'name': s['name'],
                   'isSearchable': s['is_searchable'],
                   'is_applicable': s['is_applicable']} for s in supported_entities]

            json['supportedEntities'] = se

        r = self.connection.post_request('codes', json=json)
        logger.info(f'Code {r} created.')
        return r

    def update_code(self, code_id, name=None, parent_code_id=None, description=None,
                    code_type='SourceCode', supported_entities=None):
        """
        Update a code.

        `Args:`
            code_id: int
                The code id.
            name: str
                The name of the code.
            parent_code_id: int
                A unique identifier for this code’s parent.
            description: str
                A description for this code, no longer than 200 characters.
            code_type: str
                The code type. ``Tag`` and ``SourceCode`` are valid values.
            supported_entities: list
                A list of dicts that enumerate the searchability and applicability rules of the
                code. You can find supported entities with the :meth:`code_supported_entities`

                .. highlight:: python
                .. code-block:: python

                    [
                        {
                         'name': 'Event',
                         'is_searchable': True,
                         'is_applicable': True
                        }
                        {
                         'name': 'Locations',
                         'start_time': '12-31-2018T13:00:00',
                         'end_time': '12-31-2018T14:00:00'
                        }
                    ]
        """

        post_data = {}

        if name:
            post_data['name'] = name
        if parent_code_id:
            post_data['parentCodeId'] = parent_code_id
        if code_type:
            post_data['codeType'] = code_type
        if description:
            post_data['description'] = description

        if supported_entities:

            se = [{'name': s['name'],
                   'isSearchable': s['is_searchable'],
                   'is_applicable': s['is_applicable']} for s in supported_entities]
            post_data['supportedEntities'] = se

        r = self.connection.put_request(f'codes/{code_id}', json=post_data)
        logger.info(f'Code {code_id} updated.')
        return r

    def delete_code(self, code_id):
        """
        Delete a code.

        `Args:`
            code_id: int
                The code id.
        `Returns:`
            ``None``
        """

        r = self.connection.delete_request(f'codes/{code_id}')
        logger.info(f'Code {code_id} deleted.')
        return r

    def get_code_supported_entities(self):
        """
        Get code supported entities.

        `Returns:`
            list
                A list of code supported entities.
        """

        lst = self.connection.get_request('codes/supportedEntities')
        logger.info(f'Found {len(lst)} code supported entities.')
        return lst



from parsons.notifications.smtp import SMTP
import base64
import io
import re
import unittest


class FakeConnection(object):

    def __init__(self, result_obj):
        self.result_obj = result_obj

    def sendmail(self, sender, to, message_body):
        self.result_obj.result = (sender, to, message_body)
        if 'willfail@example.com' in to:
            return {"willfail@example.com": (550, "User unknown")}

    def quit(self):
        self.result_obj.quit_ran = True


class TestSMTP(unittest.TestCase):

    def setUp(self):
        self.smtp = SMTP('fake.example.com', username='fake', password='fake')
        self.smtp.conn = FakeConnection(self)
        self.result = None
        self.quit_ran = False

    def test_send_message_simple(self):
        self.smtp.send_email('foo@example.com', 'recipient1@example.com',
                             'Simple subject', 'Fake body')
        self.assertEqual(self.result[0], 'foo@example.com')
        self.assertEqual(self.result[1], ['recipient1@example.com'])
        self.assertTrue(
            self.result[2].endswith(
                '\nto: recipient1@example.com\nfrom: foo@example.com'
                '\nsubject: Simple subject\n\nFake body'
            ))
        self.assertTrue(self.quit_ran)

    def test_send_message_html(self):
        self.smtp.send_email('foohtml@example.com', 'recipienthtml@example.com',
                             'Simple subject', 'Fake body', '<p>Really Fake html</p>')
        self.assertEqual(self.result[0], 'foohtml@example.com')
        self.assertEqual(self.result[1], ['recipienthtml@example.com'])
        self.assertRegex(self.result[2], r'<p>Really Fake html</p>\n--=======')
        self.assertRegex(self.result[2], r'\nFake body\n--======')
        self.assertRegex(self.result[2], r'ubject: Simple subject\n')
        self.assertTrue(self.quit_ran)

    def test_send_message_manualclose(self):
        smtp = SMTP('fake.example.com', username='fake', password='fake',
                    close_manually=True)
        smtp.conn = FakeConnection(self)
        smtp.send_email('foo@example.com', 'recipient1@example.com', 'Simple subject', 'Fake body')
        self.assertFalse(self.quit_ran)

    def test_send_message_files(self):
        named_file_content = 'x,y,z\n1,2,3\r\n3,4,5\r\n'
        unnamed_file_content = 'foo,bar\n1,2\r\n3,4\r\n'
        bytes_file_content = bytes(
            [71, 73, 70, 56, 57, 97, 1, 0, 1, 0, 0, 255,
             0, 44, 0, 0, 0, 0, 1, 0, 1, 0, 0, 2, 0, 59])
        named_file = io.StringIO(named_file_content)
        named_file.name = 'xyz.csv'

        bytes_file = io.BytesIO(bytes_file_content)
        bytes_file.name = 'xyz.gif'

        self.smtp.send_email(
            'foofiles@example.com', 'recipientfiles@example.com', 'Simple subject', 'Fake body',
            files=[io.StringIO(unnamed_file_content), named_file, bytes_file]
        )
        self.assertEqual(self.result[0], 'foofiles@example.com')
        self.assertEqual(self.result[1], ['recipientfiles@example.com'])
        self.assertRegex(self.result[2], r'\nFake body\n--======')
        found = re.findall(r'filename="file"\n\n([\w=/]+)\n\n--===', self.result[2])
        self.assertEqual(base64.b64decode(found[0]).decode(),
                         unnamed_file_content)
        found_named = re.findall(
            r'Content-Type: text/csv; charset="utf-8"\nMIME-Version: 1.0'
            r'\nContent-Transfer-Encoding: base64\nContent-Disposition: '
            r'attachment; filename="xyz.csv"\n\n([\w=/]+)\n\n--======',
            self.result[2])
        self.assertEqual(base64.b64decode(found_named[0]).decode(),
                         named_file_content)

        found_gif = re.findall(
            r'Content-Type: image/gif\nMIME-Version: 1.0'
            r'\nContent-Transfer-Encoding: base64\nContent-ID: <xyz.gif>'
            r'\nContent-Disposition: attachment; filename="xyz.gif"\n\n([\w=/]+)\n\n--==',
            self.result[2])
        self.assertEqual(base64.b64decode(found_gif[0]),
                         bytes_file_content)
        self.assertTrue(self.quit_ran)

    def test_send_message_partial_fail(self):
        simple_msg = self.smtp._create_message_simple(
            'foo@example.com',
            'recipient1@example.com, willfail@example.com',
            'Simple subject', 'Fake body')
        send_result = self.smtp._send_message(simple_msg)
        self.assertEqual(send_result,  {"willfail@example.com": (550, "User unknown")})



import os
from parsons.aws.s3 import S3
import time
import logging

logger = logging.getLogger(__name__)

S3_TEMP_KEY_PREFIX = "Parsons_RedshiftCopyTable"


class RedshiftCopyTable(object):

    aws_access_key_id = None
    aws_secret_access_key = None
    iam_role = None

    def __init__(self):

        pass

    def copy_statement(self, table_name, bucket, key, manifest=False,
                       data_type='csv', csv_delimiter=',', max_errors=0,
                       statupdate=True, compupdate=True, ignoreheader=1, acceptanydate=True,
                       dateformat='auto', timeformat='auto', emptyasnull=True,
                       blanksasnull=True, nullas=None, acceptinvchars=True, truncatecolumns=False,
                       specifycols=None, aws_access_key_id=None, aws_secret_access_key=None,
                       compression=None):

        # Source / Destination
        source = f's3://{bucket}/{key}'

        # Add column list for mapping or if there are fewer columns on source file
        col_list = f"({', '.join(specifycols)})" if specifycols is not None else ""

        sql = f"copy {table_name}{col_list} \nfrom '{source}' \n"

        # Generate credentials
        sql += self.get_creds(aws_access_key_id, aws_secret_access_key)

        # Other options
        if manifest:
            sql += "manifest \n"
        sql += f"maxerror {max_errors} \n"
        if statupdate:
            sql += "statupdate on\n"
        if compupdate:
            sql += "compupdate on \n"
        else:
            sql += "compupdate off \n"
        if ignoreheader:
            sql += f"ignoreheader {ignoreheader} \n"
        if acceptanydate:
            sql += "acceptanydate \n"
        sql += f"dateformat '{dateformat}' \n"
        sql += f"timeformat '{timeformat}' \n"
        if emptyasnull:
            sql += "emptyasnull \n"
        if blanksasnull:
            sql += "blanksasnull \n"
        if nullas:
            sql += f"null as {nullas}"
        if acceptinvchars:
            sql += "acceptinvchars \n"
        if truncatecolumns:
            sql += "truncatecolumns \n"

        # Data Type
        if data_type == 'csv':
            sql += f"csv delimiter '{csv_delimiter}' \n"
        else:
            raise TypeError('Invalid data type specified.')

        if compression == 'gzip':
            sql += 'gzip \n'

        sql += ';'

        return sql

    def get_creds(self, aws_access_key_id, aws_secret_access_key):

        if aws_access_key_id and aws_secret_access_key:
            # When we have credentials, then we don't need to set them again
            pass

        elif self.iam_role:
            # bail early, since the bottom is specifically formatted with creds
            return f"credentials 'aws_iam_role={self.iam_role}'\n"

        elif self.aws_access_key_id and self.aws_secret_access_key:

            aws_access_key_id = self.aws_access_key_id
            aws_secret_access_key = self.aws_secret_access_key

        elif 'AWS_ACCESS_KEY_ID' in os.environ and 'AWS_SECRET_ACCESS_KEY' in os.environ:

            aws_access_key_id = os.environ['AWS_ACCESS_KEY_ID']
            aws_secret_access_key = os.environ['AWS_SECRET_ACCESS_KEY']

        else:

            s3 = S3()
            creds = s3.aws.session.get_credentials()
            aws_access_key_id = creds.access_key
            aws_secret_access_key = creds.secret_key

        return "credentials 'aws_access_key_id={};aws_secret_access_key={}'\n".format(
            aws_access_key_id,
            aws_secret_access_key)

    def temp_s3_copy(self, tbl, aws_access_key_id=None, aws_secret_access_key=None):

        if not self.s3_temp_bucket:
            raise KeyError(("Missing S3_TEMP_BUCKET, needed for transferring data to Redshift. "
                            "Must be specified as env vars or kwargs"
                            ))

        self.s3 = S3(aws_access_key_id=aws_access_key_id,
                     aws_secret_access_key=aws_secret_access_key)

        hashed_name = hash(time.time())
        key = f"{S3_TEMP_KEY_PREFIX}/{hashed_name}.csv.gz"

        # Convert table to compressed CSV file, to optimize the transfers to S3 and to
        # Redshift.
        local_path = tbl.to_csv(temp_file_compression='gzip')
        # Copy table to bucket
        self.s3.put_file(self.s3_temp_bucket, key, local_path)

        return key

    def temp_s3_delete(self, key):

        if key:
            self.s3.remove_file(self.s3_temp_bucket, key)



class RedshiftSchema(object):

    def schema_exists(self, schema):
        sql = f"select * from pg_namespace where nspname = '{schema}'"
        res = self.query(sql)
        return res.num_rows > 0

    def create_schema_with_permissions(self, schema, group=None):
        """
        Creates a Redshift schema (if it doesn't already exist), and grants usage permissions to
        a Redshift group (if specified).

        `Args:`
            schema: str
                The schema name
            group: str
                The Redshift group name
            type: str
                The type of permissions to grant. Supports `select`, `all`, etc. (For
                full list, see the
                `Redshift GRANT docs <https://docs.aws.amazon.com/redshift/latest/dg/r_GRANT.html>`_)
        """ # noqa: E501,E261

        if not self.schema_exists(schema):
            self.query(f"create schema {schema}")
            self.query(f"grant usage on schema {schema} to group {group}")

    def grant_schema_permissions(self, schema, group, permissions_type='select'):
        """
        Grants a Redshift group permissions to all tables within an existing schema.

        `Args:`
            schema: str
                The schema name
            group: str
                The Redshift group name
            type: str
                The type of permissions to grant. Supports `select`, `all`, etc. (For
                full list, see the
                `Redshift GRANT docs <https://docs.aws.amazon.com/redshift/latest/dg/r_GRANT.html>`_)
        """ # noqa: E501,E261

        sql = f"""
            grant usage on schema {schema} to group {group};
            grant {permissions_type} on all tables in schema {schema} to group {group};
        """
        self.query(sql)



import unittest
import os
import requests_mock
from parsons.ngpvan.van import VAN
from requests.exceptions import HTTPError
from test.test_van.responses_people import *

os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestNGPVAN(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    @requests_mock.Mocker()
    def test_find_person(self, m):

        m.post(self.van.connection.uri + 'people/find', json=find_people_response, status_code=200)

        person = self.van.find_person(first_name='Bob', last_name='Smith', phone=4142020792)

        self.assertEqual(person, find_people_response)

    @requests_mock.Mocker()
    def test_find_person_json(self, m):

        json = {
            "firstName": "Bob",
            "lastName": "Smith",
            "phones": [{
                "phoneNumber": 4142020792
            }]
        }

        m.post(self.van.connection.uri + 'people/find', json=find_people_response, status_code=200)

        person = self.van.find_person_json(match_json=json)

        self.assertEqual(person, find_people_response)

    def test_upsert_person(self):

        pass

    def test_upsert_person_json(self):

        pass

    def test_update_person(self):

        pass

    def test_update_person_json(self):

        pass

    def test_people_search(self):

        # Already tested as part of upsert and find person methods
        pass

    def test_valid_search(self):

        # Fails with FN / LN Only
        self.assertRaises(ValueError, self.van._valid_search, 'Barack',
                          'Obama', None, None, None, None, None)

        # Fails with only Zip
        self.assertRaises(ValueError, self.van._valid_search, 'Barack',
                          'Obama', None, None, None, None, 60622)

        # Fails with no street number
        self.assertRaises(ValueError, self.van._valid_search, 'Barack',
                          'Obama', None, None, None, 'Pennsylvania Ave', None)

        # Successful with FN/LN/Email
        self.van._valid_search('Barack', 'Obama', 'barack@email.com', None, None, None,
                               None)

        # Successful with FN/LN/DOB/ZIP
        self.van._valid_search('Barack', 'Obama', 'barack@email.com', None, '2000-01-01',
                               None, 20009)

        # Successful with FN/LN/Phone
        self.van._valid_search('Barack', 'Obama', None, 2024291000, None, None,
                               None)

    @requests_mock.Mocker()
    def test_get_person(self, m):

        json = get_person_response

        # Test works with external ID
        m.get(self.van.connection.uri + f'people/DWID:15406767', json=json)
        person = self.van.get_person('15406767', id_type='DWID')
        self.assertEqual(get_person_response, person)

        # Test works with vanid
        m.get(self.van.connection.uri + f'people/19722445', json=json)
        person = self.van.get_person('19722445')
        self.assertEqual(get_person_response, person)

    @requests_mock.Mocker()
    def test_apply_canvass_result(self, m):

        # Test a valid attempt
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=204)
        self.van.apply_canvass_result(2335282, 18)


        # Test a bad result code
        json = {'errors':
                [{'code': 'INVALID_PARAMETER',
                    'text': "'resultCodeId' must be a valid result code in the current context.",
                    'properties': ['resultCodeId']}
                 ]}
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses',
               json=json, status_code=400)
        self.assertRaises(HTTPError, self.van.apply_canvass_result, 2335282, 0)

        # Test a bad vanid
        json = {'errors':
                [{'code': 'INTERNAL_SERVER_ERROR',
                  'text': 'An unknown error occurred',
                  'referenceCode': '88A111-E2FF8'}
                 ]}
        m.post(self.van.connection.uri + f'people/0/canvassResponses', json=json, status_code=400)
        self.assertRaises(HTTPError, self.van.apply_canvass_result, 0, 18)

        # Test a good dwid
        m.post(self.van.connection.uri + f'people/DWID:2335282/canvassResponses', status_code=204)
        self.van.apply_canvass_result(2335282, 18, id_type='DWID')

    @requests_mock.Mocker()
    def test_apply_survey_question(self, m):

        # Test valid survey question
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=204)
        self.van.apply_survey_response(2335282, 351006, 1443891)

        # Test bad survey response id
        json = {'errors':
                [{'code': 'INVALID_PARAMETER',
                  'text': "'surveyResponseId' must be a valid Response to the given Survey Question.",
                  'properties': ['responses[0].surveyResponseId']}
                 ]}
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=400)
        self.assertRaises(HTTPError, self.van.apply_survey_response, 2335282, 0, 1443891)

        # Test bad survey question id
        json = {'errors':
                [{'code': 'INVALID_PARAMETER',
                  'text': "'surveyQuestionId' must be a valid Survey Question that is available in the current context.",
                  'properties': ['responses[0].surveyQuestionId']}]}
        m.post(self.van.connection.uri + f'people/2335282/canvassResponses', status_code=400)
        self.assertRaises(HTTPError, self.van.apply_survey_response, 2335282, 351006, 0)

    def test_toggle_volunteer_action(self):

        pass

    def test_apply_response(self):

        pass

    @requests_mock.Mocker()
    def test_create_relationship(self, m):

        relationship_id = 12
        bad_vanid_1 = 99999
        good_vanid_1 = 12345
        vanid_2 = 54321

        # Bad request
        m.post(self.van.connection.uri + "people/{}/relationships".format(bad_vanid_1),
               status_code=404)

        # Good request
        m.post(self.van.connection.uri + "people/{}/relationships".format(good_vanid_1),
               status_code=204)

        # Test bad input
        self.assertRaises(HTTPError, self.van.create_relationship, bad_vanid_1, vanid_2, relationship_id)
        self.assertRaises(HTTPError, self.van.create_relationship, bad_vanid_1, vanid_2, relationship_id)

        self.van.create_relationship(good_vanid_1, vanid_2, relationship_id)

    @requests_mock.Mocker()
    def test_apply_person_code(self, m):

        vanid = 999
        code_id = 888

        # Test good request
        m.post(self.van.connection.uri + f"people/{vanid}/codes", status_code=204)
        self.van.apply_person_code(vanid, code_id)

        # Test bad request
        m.post(self.van.connection.uri + f"people/{vanid}/codes", status_code=404)
        self.assertRaises(HTTPError, self.van.apply_person_code, vanid, code_id)



import unittest
import os
import requests_mock
from parsons.ngpvan.van import VAN

os.environ['VAN_API_KEY'] = 'SOME_KEY'


class TestBulkImport(unittest.TestCase):

    def setUp(self):

        self.van = VAN(os.environ['VAN_API_KEY'], db="MyVoters", raise_for_status=False)

    def tearDown(self):

        pass

    @requests_mock.Mocker()
    def test_get_bulk_import_resources(self, m):

        json = ['Contacts', 'Contributions', 'ActivistCodes', 'ContactsActivistCodes']

        m.get(self.van.connection.uri + 'bulkImportJobs/resources', json=json)

        self.assertEqual(self.van.get_bulk_import_resources(), json)



import unittest
import os
from parsons.civis import CivisClient
from parsons.etl.table import Table

# from . import scratch_creds

@unittest.skipIf(not os.environ.get('LIVE_TEST'), 'Skipping because not running live test')
class TestCivisClient(unittest.TestCase):

    def setUp(self):

        self.civis = CivisClient()

        # Create a schema, create a table, create a view
        setup_sql = """
                    drop schema if exists test_parsons cascade;
                    create schema test_parsons;
                    """

        self.lst_dicts = [{'first': 'Bob', 'last': 'Smith'}]
        self.tbl = Table(self.lst_dicts)

        self.civis.query(setup_sql)

    def tearDown(self):

        # Drop the view, the table and the schema
        teardown_sql = """
                       drop schema if exists test_parsons cascade;
                       """

        self.civis.query(teardown_sql)

    def test_table_import_query(self):

        # Test that a good table imports correctly
        self.civis.table_import(self.tbl, 'test_parsons.test_table')

    def test_query(self):

        # Test that queries match
        self.civis.table_import(self.tbl, 'test_parsons.test_table')
        tbl = self.civis.query("SELECT COUNT(*) FROM test_parsons.test_table")
        self.assertEqual(tbl[0]['count'], '1')

    def test_to_civis(self):

        # Test that the to_civis() method works too
        self.tbl.to_civis('test_parsons.test_table')
